from fhir.resources.timing import Timing, TimingRepeat
from medical.models import Service
from fhir.resources.activitydefinition import ActivityDefinition
from fhir.resources.extension import Extension
from fhir.resources.money import Money
from fhir.resources.usagecontext import UsageContext
from fhir.resources.codeableconcept import CodeableConcept

from api_fhir_r4.configurations import GeneralConfiguration
from api_fhir_r4.converters import R4IdentifierConfig, BaseFHIRConverter, ReferenceConverterMixin
from django.utils.translation import gettext

from api_fhir_r4.mapping.activityDefinitionMapping import ServiceTypeMapping, UseContextMapping, VenueMapping, \
    WorkflowMapping, ServiceLevelMapping
from api_fhir_r4.mapping.patientMapping import PatientCategoryMapping
from api_fhir_r4.utils import DbManagerUtils, TimeUtils
import core


class ActivityDefinitionConverter(BaseFHIRConverter, ReferenceConverterMixin):

    @classmethod
    def to_fhir_obj(cls, imis_activity_definition, reference_type=ReferenceConverterMixin.UUID_REFERENCE_TYPE):
        fhir_activity_definition = ActivityDefinition.construct()
        # first to construct is status - obligatory fields
        cls.build_fhir_status(fhir_activity_definition, imis_activity_definition)
        cls.build_fhir_pk(fhir_activity_definition, imis_activity_definition, reference_type)
        cls.build_fhir_identifiers(fhir_activity_definition, imis_activity_definition)
        cls.build_fhir_date(fhir_activity_definition, imis_activity_definition)
        cls.build_fhir_name(fhir_activity_definition, imis_activity_definition)
        cls.build_fhir_title(fhir_activity_definition, imis_activity_definition)
        cls.build_fhir_use_context(fhir_activity_definition, imis_activity_definition)
        cls.build_fhir_topic(fhir_activity_definition, imis_activity_definition)
        cls.build_fhir_timing(fhir_activity_definition, imis_activity_definition)
        cls.build_fhir_activity_definition_extension(fhir_activity_definition, imis_activity_definition)
        cls.build_fhir_level(fhir_activity_definition, imis_activity_definition)
        return fhir_activity_definition

    @classmethod
    def get_fhir_code_identifier_type(cls):
        return R4IdentifierConfig.get_fhir_generic_type_code()

    @classmethod
    def to_imis_obj(cls, fhir_activity_definition, audit_user_id):
        errors = []
        fhir_activity_definition = ActivityDefinition(**fhir_activity_definition)
        imis_activity_definition = Service()
        imis_activity_definition.audit_user_id = audit_user_id
        cls.build_imis_identifier(imis_activity_definition, fhir_activity_definition, errors)
        cls.build_imis_validity_from(imis_activity_definition, fhir_activity_definition, errors)
        cls.build_imis_price(imis_activity_definition, fhir_activity_definition, errors)
        cls.build_imis_frequency(imis_activity_definition, fhir_activity_definition, errors)
        cls.build_imis_serv_code(imis_activity_definition, fhir_activity_definition, errors)
        cls.build_imis_serv_name(imis_activity_definition, fhir_activity_definition, errors)
        cls.build_imis_serv_type(imis_activity_definition, fhir_activity_definition, errors)
        cls.build_imis_serv_pat_cat(imis_activity_definition, fhir_activity_definition, errors)
        cls.build_imis_serv_category(imis_activity_definition, fhir_activity_definition, errors)
        cls.build_imis_serv_care_type(imis_activity_definition, fhir_activity_definition, errors)
        cls.build_imis_level(imis_activity_definition, fhir_activity_definition, errors)
        cls.check_errors(errors)
        return imis_activity_definition

    @classmethod
    def get_reference_obj_uuid(cls, imis_activity_definition: Service):
        return imis_activity_definition.uuid

    @classmethod
    def get_reference_obj_id(cls, imis_activity_definition: Service):
        return imis_activity_definition.id

    @classmethod
    def get_reference_obj_code(cls, imis_activity_definition: Service):
        return imis_activity_definition.code

    @classmethod
    def get_fhir_resource_type(cls):
        return ActivityDefinition

    @classmethod
    def get_imis_obj_by_fhir_reference(cls, reference, errors=None):
        imis_activity_definition_code = cls.get_resource_id_from_reference(reference)
        return DbManagerUtils.get_object_or_none(Service, uuid=imis_activity_definition_code)

    @classmethod
    def build_fhir_identifiers(cls, fhir_activity_definition, imis_activity_definition):
        identifiers = []
        cls.build_all_identifiers(identifiers, imis_activity_definition)
        fhir_activity_definition.identifier = identifiers

    @classmethod
    def build_imis_price(cls, imis_activity_definition, fhir_activity_definition, errors):
        extension_base_url = f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/unit-price"
        extension = cls.get_fhir_extension_by_url(fhir_activity_definition.extension,
                                                  extension_base_url)

        if extension and hasattr(extension, "valueMoney"):
            imis_activity_definition.price = extension.valueMoney.value

    @classmethod
    def build_imis_frequency(cls, imis_activity_definition, fhir_activity_definition, errors):
        value = None
        try:
            if hasattr(fhir_activity_definition, "timingTiming"):
                value = fhir_activity_definition.timingTiming.repeat.period
        except AttributeError:
            errors.append(gettext('Invalid activity definition `timingTiming` attribute'))
        finally:
            imis_activity_definition.frequency = value

    @classmethod
    def build_imis_identifier(cls, imis_activity_definition, fhir_activity_definition, errors):
        value = cls.get_fhir_identifier_by_code(fhir_activity_definition.identifier,
                                                R4IdentifierConfig.get_fhir_uuid_type_code())
        if value:
            imis_activity_definition.uuid = value

        value = cls.get_fhir_identifier_by_code(fhir_activity_definition.identifier,
                                                ActivityDefinitionConverter.get_fhir_code_identifier_type())
        if value:
            imis_activity_definition.code = value

        cls.valid_condition(not imis_activity_definition.code,
                            gettext('Missing the service code'), errors)

    @classmethod
    def build_fhir_status(cls, fhir_activity_definition, imis_activity_definition):
        fhir_activity_definition.status = "active"

    @classmethod
    def build_fhir_date(cls, fhir_activity_definition, imis_activity_definition):
        fhir_activity_definition.date = imis_activity_definition.validity_from.isoformat()

    @classmethod
    def build_imis_validity_from(cls, imis_activity_definition, fhir_activity_definition, errors):
        validity_from = fhir_activity_definition.date
        if not cls.valid_condition(not validity_from,
                                   gettext('Missing activity definition `validity from` attribute'), errors):
            imis_activity_definition.validity_from = TimeUtils.str_iso_to_date(validity_from)

    @classmethod
    def build_fhir_name(cls, fhir_activity_definition, imis_activity_definition):
        fhir_activity_definition.name = imis_activity_definition.code

    @classmethod
    def build_imis_serv_code(cls, imis_activity_definition, fhir_activity_definition, errors):
        serv_code = fhir_activity_definition.name
        if not cls.valid_condition(not serv_code,
                                   gettext('Missing activity definition `serv code` attribute'), errors):
            imis_activity_definition.code = serv_code

    @classmethod
    def build_fhir_title(cls, fhir_activity_definition, imis_activity_definition):
        fhir_activity_definition.title = imis_activity_definition.name

    @classmethod
    def build_imis_serv_name(cls, imis_activity_definition, fhir_activity_definition, errors):
        serv_name = fhir_activity_definition.title
        if not cls.valid_condition(not serv_name,
                                   gettext('Missing activity definition `serv name` attribute'), errors):
            imis_activity_definition.name = serv_name

    @classmethod
    def build_imis_serv_pat_cat(cls, imis_activity_definition, fhir_activity_definition, errors):
        use_context_codes = ["gender", "age"]
        pat_cat_flag = 0
        try:
            for use_context_code in use_context_codes:
                use_context = cls.get_use_context_by_code(fhir_activity_definition.useContext, use_context_code)
                if use_context and len(use_context.valueCodeableConcept.coding) > 0:
                    for coding in use_context.valueCodeableConcept.coding:
                        pat_cat_flag = pat_cat_flag | PatientCategoryMapping.imis_patient_category_flags[coding.code]
        except AttributeError or TypeError or KeyError:
            errors.append(gettext('Invalid activity definition `use context - age/gender` attribute'))
        finally:
            imis_activity_definition.patient_category = pat_cat_flag

    @classmethod
    def build_imis_serv_category(cls, imis_activity_definition, fhir_activity_definition, errors):
        use_context = cls.get_use_context_by_code(fhir_activity_definition.useContext, "workflow")
        service_category = None
        try:
            if use_context and len(use_context.valueCodeableConcept.coding) > 0:
                service_category = use_context.valueCodeableConcept.coding[0].code
        except AttributeError or TypeError or KeyError:
            errors.append(gettext('Invalid activity definition `use context - workflow` attribute'))
        finally:
            imis_activity_definition.category = service_category

    @classmethod
    def build_imis_serv_care_type(cls, imis_activity_definition, fhir_activity_definition, errors):
        use_context = cls.get_use_context_by_code(fhir_activity_definition.useContext, "venue")
        care_type = None
        try:
            if not use_context or len(use_context.valueCodeableConcept.coding) == 0:
                errors.append(gettext('Missing activity definition `use context - venue` attribute'))
            else:
                for coding in use_context.valueCodeableConcept.coding:
                    if coding.code not in VenueMapping.imis_venue_coding.keys():
                        raise AttributeError
                    if not care_type:
                        care_type = VenueMapping.imis_venue_coding[coding.code]
                    elif care_type != coding.code:
                        care_type = Service.CARE_TYPE_BOTH
        except AttributeError or TypeError or KeyError:
            errors.append(gettext('Invalid activity definition `use context - venue` attribute'))
        finally:
            imis_activity_definition.care_type = care_type

    @classmethod
    def build_imis_serv_type(cls, imis_activity_definition, fhir_activity_definition, errors):
        service_type = None
        try:
            if hasattr(fhir_activity_definition, "topic"):
                service_type = cls.get_first_coding_from_codeable_concept(fhir_activity_definition.topic[0]).code
                if service_type not in ServiceTypeMapping.fhir_service_type_coding.keys():
                    raise AttributeError
        except AttributeError:
            errors.append(gettext('Invalid activity definition `topic` attribute'))
        finally:
            imis_activity_definition.type = service_type

    @classmethod
    def build_fhir_activity_definition_extension(cls, fhir_activity_definition, imis_activity_definition):
        cls.build_fhir_unit_price(fhir_activity_definition, imis_activity_definition)
        return fhir_activity_definition

    @classmethod
    def build_fhir_unit_price(cls, fhir_activity_definition, imis_activity_definition):
        unit_price = cls.build_fhir_unit_price_extension(imis_activity_definition.price)
        if type(fhir_activity_definition.extension) is not list:
            fhir_activity_definition.extension = [unit_price]
        else:
            fhir_activity_definition.extension.append(unit_price)

    @classmethod
    def build_fhir_unit_price_extension(cls, value):
        extension = Extension.construct()
        money = Money.construct()
        extension.url = f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/unit-price"
        extension.valueMoney = money
        extension.valueMoney.value = value
        if hasattr(core, 'currency'):
            extension.valueMoney.currency = core.currency
        return extension

    @classmethod
    def build_fhir_topic(cls, fhir_activity_definition, imis_activity_definition):
        topic = [cls.build_codeable_concept_from_coding(cls.build_fhir_topic_coding(imis_activity_definition.type))]
        fhir_activity_definition.topic = topic

    @classmethod
    def build_fhir_use_context(cls, fhir_activity_definition, imis_activity_definition):
        usage = []
        if imis_activity_definition.patient_category:
            usage.append(cls.build_fhir_gender_usage_context(imis_activity_definition))
            usage.append(cls.build_fhir_age_usage_context(imis_activity_definition))

        if imis_activity_definition.category and not imis_activity_definition.category.isspace():
            usage.append(cls.build_fhir_workflow_usage_context(imis_activity_definition))

        usage.append(cls.build_fhir_venue_usage_context(imis_activity_definition))

        usage = [usage_item for usage_item in usage if len(usage_item.valueCodeableConcept.coding) > 0]
        fhir_activity_definition.useContext = usage

    @classmethod
    def build_fhir_usage_context(cls, code, codeable_concept):
        usage_context = UsageContext.construct()
        usage_context.valueCodeableConcept = codeable_concept
        usage_context.code = code
        return usage_context

    @classmethod
    def build_fhir_gender_usage_context(cls, imis_activity_definition):
        return cls.build_fhir_usage_context(cls.build_fhir_usage_context_coding("gender"),
                                            cls.build_fhir_gender(imis_activity_definition))

    @classmethod
    def build_fhir_age_usage_context(cls, imis_activity_definition):
        return cls.build_fhir_usage_context(cls.build_fhir_usage_context_coding("age"),
                                            cls.build_fhir_age(imis_activity_definition))

    @classmethod
    def build_fhir_venue_usage_context(cls, imis_activity_definition):
        return cls.build_fhir_usage_context(cls.build_fhir_usage_context_coding("venue"),
                                            cls.build_fhir_venue(imis_activity_definition))

    @classmethod
    def build_fhir_workflow_usage_context(cls, imis_activity_definition):
        return cls.build_fhir_usage_context(cls.build_fhir_usage_context_coding("workflow"),
                                            cls.build_fhir_workflow(imis_activity_definition))

    @classmethod
    def build_fhir_workflow(cls, imis_activity_definition):
        codeable_concept = CodeableConcept.construct()
        codeable_concept.coding = []

        codeable_concept.coding.append(cls.build_fhir_workflow_coding(imis_activity_definition.category))
        codeable_concept.text = codeable_concept.coding[0].display

        return codeable_concept

    @classmethod
    def build_fhir_venue(cls, imis_activity_definition):
        codeable_concept = CodeableConcept.construct()
        codeable_concept.coding = []

        ambulatory_venue = imis_activity_definition.care_type in [Service.CARE_TYPE_OUT_PATIENT, Service.CARE_TYPE_BOTH]
        if ambulatory_venue:
            ambulatory_coding = cls.build_fhir_venue_coding(Service.CARE_TYPE_OUT_PATIENT)
            codeable_concept.coding.append(ambulatory_coding)

        imp_venue = imis_activity_definition.care_type in [Service.CARE_TYPE_IN_PATIENT, Service.CARE_TYPE_BOTH]
        if imp_venue:
            imp_coding = cls.build_fhir_venue_coding(Service.CARE_TYPE_IN_PATIENT)
            codeable_concept.coding.append(imp_coding)

        codeable_concept.text = " or ".join([coding.display for coding in codeable_concept.coding])

        return codeable_concept

    @classmethod
    def build_fhir_gender(cls, imis_activity_definition):
        codeable_concept = CodeableConcept.construct()
        codeable_concept.coding = []

        male_flag = PatientCategoryMapping.imis_patient_category_flags["male"]
        if imis_activity_definition.patient_category & male_flag:
            male_coding = cls.build_fhir_patient_category_coding("male")
            codeable_concept.coding.append(male_coding)

        female_flag = PatientCategoryMapping.imis_patient_category_flags["female"]
        if imis_activity_definition.patient_category & female_flag:
            female_coding = cls.build_fhir_patient_category_coding("female")
            codeable_concept.coding.append(female_coding)

        codeable_concept.text = " or ".join([coding.display for coding in codeable_concept.coding])

        return codeable_concept

    @classmethod
    def build_fhir_age(cls, imis_activity_definition):
        codeable_concept = CodeableConcept.construct()
        codeable_concept.coding = []

        adult_flag = PatientCategoryMapping.imis_patient_category_flags["adult"]
        if imis_activity_definition.patient_category & adult_flag:
            adult_coding = cls.build_fhir_patient_category_coding("adult")
            codeable_concept.coding.append(adult_coding)

        child_flag = PatientCategoryMapping.imis_patient_category_flags["child"]
        if imis_activity_definition.patient_category & child_flag:
            child_coding = cls.build_fhir_patient_category_coding("child")
            codeable_concept.coding.append(child_coding)

        codeable_concept.text = " or ".join([coding.display for coding in codeable_concept.coding])

        return codeable_concept

    @classmethod
    def build_fhir_usage_context_coding(cls, usage_context):
        return cls.build_fhir_mapped_coding(UseContextMapping.fhir_use_context_coding[usage_context])

    @classmethod
    def build_fhir_patient_category_coding(cls, category):
        return cls.build_fhir_mapped_coding(PatientCategoryMapping.fhir_patient_category_coding[category])

    @classmethod
    def build_fhir_venue_coding(cls, venue):
        return cls.build_fhir_mapped_coding(VenueMapping.fhir_venue_coding[venue])

    @classmethod
    def build_fhir_workflow_coding(cls, workflow):
        return cls.build_fhir_mapped_coding(WorkflowMapping.fhir_workflow_coding[workflow])

    @classmethod
    def build_fhir_topic_coding(cls, topic):
        return cls.build_fhir_mapped_coding(ServiceTypeMapping.fhir_service_type_coding[topic])

    @classmethod
    def build_fhir_timing(cls, fhir_activity_definition, imis_activity_definition):
        timing = Timing.construct()
        timing_repeat = TimingRepeat.construct()

        timing_repeat.frequency = 1
        timing_repeat.period = imis_activity_definition.frequency
        timing_repeat.periodUnit = "d"

        timing.repeat = timing_repeat
        fhir_activity_definition.timingTiming = timing

    @classmethod
    def build_fhir_level(cls, fhir_activity_definition: ActivityDefinition, imis_activity_definition: Service):
        imis_level = imis_activity_definition.level
        coding_data = ServiceLevelMapping.fhir_service_level_coding.get(imis_level, None)
        if coding_data:
            coding = cls.build_fhir_mapped_coding(coding_data)
            extension = Extension(
                url=f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/activity-definition-level",
                valueCodeableConcept=CodeableConcept(
                    coding=[coding],
                    text=coding.display
                )
            )
            if isinstance(fhir_activity_definition.extension, list):
                fhir_activity_definition.extension.append(extension)
            else:
                fhir_activity_definition.extension = [extension]

    @classmethod
    def build_imis_level(cls, imis_service: Service, fhir_service: ActivityDefinition, errors: list):
        level_ext = next(
            (ext for ext in fhir_service.extension if ext.url.lower().endswith('activity-definition-level')),
            None
        )

        if level_ext:
            code = level_ext.valueCodeableConcept.coding[0].code
            valid_codes = ServiceLevelMapping.fhir_service_level_coding.keys()

            if code not in valid_codes:
                errors.append(
                    F"Invalid code system {code} for ActivityDefinition Level, "
                    F"Valid codes are: {valid_codes}"
                )

            imis_service.level = code



import json
import os

from django.utils.translation import gettext as _
from fhir.resources.group import Group
from rest_framework import status
from rest_framework.test import APITestCase

from api_fhir_r4.configurations import GeneralConfiguration
from api_fhir_r4.tests import GenericFhirAPITestMixin, FhirApiReadTestMixin, GroupTestMixin
from api_fhir_r4.tests.mixin.logInMixin import LogInMixin
from insuree.test_helpers import *


class GroupAPITests(GenericFhirAPITestMixin, FhirApiReadTestMixin, APITestCase, LogInMixin):
    base_url = GeneralConfiguration.get_base_url() + 'Group/'
    _test_json_path = "/test/test_group.json"
    _TEST_INSUREE_CHFID = "TestChfId1"
    _TEST_INSUREE_LAST_NAME = "Test"
    _TEST_INSUREE_OTHER_NAMES = "TestInsuree"
    _TEST_POVERTY_STATUS = True
    _TEST_INSUREE_CHFID_NOT_EXIST = "NotExistedCHF"

    _test_json_path_credentials = "/tests/test/test_login.json"
    _test_request_data_credentials = None

    def setUp(self):
        super(GroupAPITests, self).setUp()
        dir_path = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
        json_representation = open(dir_path + self._test_json_path_credentials).read()
        self._test_request_data_credentials = json.loads(json_representation)
        self.get_or_create_user_api()

    def verify_updated_obj(self, updated_obj):
        self.assertTrue(isinstance(updated_obj, Group))
        poverty_data = None
        for extension in updated_obj.extension:
            if "group-poverty-status" in extension.url:
                poverty_data = extension
        self.assertEqual(self._TEST_POVERTY_STATUS, poverty_data.valueBoolean)

    def update_resource(self, data):
        for extension in data["extension"]:
            if "group-poverty-status" in extension["url"]:
                extension["valueBoolean"] = self._TEST_POVERTY_STATUS

    def create_dependencies(self):
        insuree = create_test_insuree(
            with_family=False,
            custom_props=
            {
                "chf_id": self._TEST_INSUREE_CHFID,
                "last_name": self._TEST_INSUREE_LAST_NAME,
                "other_names": self._TEST_INSUREE_OTHER_NAMES,
            }
        )
        imis_location = GroupTestMixin().create_mocked_location()
        imis_location.save()

    def update_payload_no_extensions(self, data):
        data["extension"] = []
        return data

    def update_payload_no_such_chf_id(self, data):
        for member in data["member"]:
            member["entity"]["reference"] = f"Patient/{self._TEST_INSUREE_CHFID_NOT_EXIST}"
        return data

    def update_payload_remove_chf_id_from_it(self, data):
        for member in data["member"]:
            member["entity"].pop("reference")
        return data

    def test_post_should_create_correctly(self):
        self.create_dependencies()
        response = self.client.post(
            GeneralConfiguration.get_base_url() + 'login/', data=self._test_request_data_credentials, format='json'
        )
        response_json = response.json()
        token = response_json["token"]
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        headers = {
            "Content-Type": "application/json",
            'HTTP_AUTHORIZATION': f"Bearer {token}"
        }
        response = self.client.post(self.base_url, data=self._test_request_data, format='json', **headers)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertIsNotNone(response.content)

    def test_post_should_raise_error_no_extensions(self):
        self.login()
        self.create_dependencies()
        modified_payload = self.update_payload_no_extensions(data=self._test_request_data)
        response = self.client.post(self.base_url, data=modified_payload, format='json')
        response_json = response.json()
        self.assertEqual(
            response_json["issue"][0]["details"]["text"],
            _("At least one extension with address is required")
        )

    def test_post_should_raise_error_no_such_chf_id(self):
        self.login()
        self.create_dependencies()
        modified_payload = self.update_payload_no_such_chf_id(data=self._test_request_data)
        response = self.client.post(self.base_url, data=modified_payload, format='json')
        response_json = response.json()
        self.assertEqual(
            response_json["issue"][0]["details"]["text"],
            _('Such insuree %(chf_id)s does not exist') % {'chf_id': self._TEST_INSUREE_CHFID_NOT_EXIST}
        )

    def test_post_should_raise_error_no_chf_id_in_payload(self):
        self.login()
        self.create_dependencies()
        modified_payload = self.update_payload_remove_chf_id_from_it(data=self._test_request_data)
        response = self.client.post(self.base_url, data=modified_payload, format='json')
        response_json = response.json()
        self.assertEqual(
            response_json["issue"][0]["details"]["text"],
            _("Family Group FHIR without code - this field is obligatory")
        )



import copy
from functools import lru_cache
from urllib.parse import urljoin

from api_fhir_r4.configurations import GeneralConfiguration, R4OrganisationConfig
from api_fhir_r4.models.imisModelEnums import ContactPointSystem
from policyholder.apps import PolicyholderConfig
from policyholder.services import PolicyHolderActivity, PolicyHolderLegalForm


class HealthFacilityOrganizationTypeMapping:
    LEGAL_FORM_CODE = 'D'
    LEGAL_FORM_DISPLAY = 'District organization'
    LEGAL_FORM_SYSTEM = 'CodeSystem/organization-legal-form'
    LEGAL_FORM_URL = f'{GeneralConfiguration.get_system_base_url()}StructureDefinition/organization-legal-form'

    LEGAL_FORM_MAPPING = {
        'C': 'Charity',
        'D': 'District Organization',
        'G': 'Government',
        'P': 'Private organization'
    }

    ORGANIZATION_TYPE = 'prov'

    EMAIL_CONTACT_POINT_SYSTEM = ContactPointSystem.EMAIL
    PHONE_CONTACT_POINT_SYSTEM = ContactPointSystem.PHONE
    FAX_CONTACT_POINT_SYSTEM = ContactPointSystem.FAX

    ADDRESS_LOCATION_REFERENCE_URL = \
        f'{GeneralConfiguration.get_system_base_url()}/StructureDefinition/address-location-reference'

    CONTRACT_PURPOSE = {
        'code': 'PAYOR',
        'system': 'http://terminology.hl7.org/CodeSystem/contactentity-type'
    }

    LEVEL_DISPLAY_MAPPING = {
        'D': 'Dispensary',
        'C': 'Health Centre',
        'H': 'Hospital'
    }

    LEVEL_SYSTEM = f'{GeneralConfiguration.get_system_base_url()}/CodeSystem/organization-hf-level'

    TYPE_SYSTEM = f'{GeneralConfiguration.get_system_base_url()}/CodeSystem/organization-hf-care-type'

    TYPE_DISPLAY_MAPPING = {
        'O': 'Out-patient',
        'I': 'In-patient',
        'B': 'Both'
    }


class PolicyHolderConfigMapping:
    @classmethod
    def _get_system(cls):
        raise NotImplementedError('_get_system() not implemented')

    @classmethod
    def _get_config_mapping(cls):
        raise NotImplementedError('_get_config_mapping() not implemented')

    @classmethod
    def fhir_ph_code_system(cls, code):
        @lru_cache(maxsize=None)
        def __code_system_dict():
            system = urljoin(
                GeneralConfiguration.get_system_base_url(),
                cls._get_system()
            )
            dict_ = {}
            for activity in copy.deepcopy(cls._get_config_mapping()):
                activity['system'] = system
                dict_[int(activity['code'])] = activity
            return dict_

        return __code_system_dict()[code]


class PolicyHolderOrganisationLegalFormMapping(PolicyHolderConfigMapping):
    @classmethod
    def _get_system(cls):
        return R4OrganisationConfig.get_fhir_ph_organisation_legal_form_code_system()

    @classmethod
    def _get_config_mapping(cls):
        return PolicyholderConfig.policyholder_legal_form


class PolicyHolderOrganisationActivityMapping(PolicyHolderConfigMapping):
    @classmethod
    def _get_system(cls):
        return R4OrganisationConfig.get_fhir_ph_organisation_activity_code_system()

    @classmethod
    def _get_config_mapping(cls):
        return PolicyholderConfig.policyholder_activity



from rest_framework.test import APITestCase

from api_fhir_r4.configurations import GeneralConfiguration
from api_fhir_r4.tests import GenericFhirAPITestMixin, FhirApiReadTestMixin
from api_fhir_r4.tests.mixin.invoiceTestMixin import InvoiceTestMixin
from api_fhir_r4.tests.mixin.logInMixin import LogInMixin
from insuree.models import Insuree


class InvoiceAPITests(GenericFhirAPITestMixin, FhirApiReadTestMixin, APITestCase, LogInMixin, InvoiceTestMixin):
    base_url = GeneralConfiguration.get_base_url() + 'Invoice/?resourceType=invoice'
    _test_json_path = "/test/test_invoice.json"

    def setUp(self):
        super(InvoiceAPITests, self).setUp()
        user = self.get_or_create_user_api()
        self.imis_invoice, self.imis_invoice_line_item = self.create_test_imis_instance()
        self.imis_invoice.thirdparty = Insuree.objects.first()
        self.imis_invoice.save(username=user.username)
        self.imis_invoice_line_item.save(username=user.username)



from rest_framework import mixins
from rest_framework.viewsets import GenericViewSet

from api_fhir_r4.mixins import MultiIdentifierRetrieverMixin
from api_fhir_r4.model_retrievers import UUIDIdentifierModelRetriever, CodeIdentifierModelRetriever
from api_fhir_r4.permissions import FHIRApiClaimPermissions
from api_fhir_r4.serializers import ClaimResponseSerializer
from api_fhir_r4.views.fhir.base import BaseFHIRView
from api_fhir_r4.views.filters import ValidityFromRequestParameterFilter
from claim.models import Claim


class ClaimResponseViewSet(BaseFHIRView, MultiIdentifierRetrieverMixin, mixins.ListModelMixin, GenericViewSet,
                           mixins.UpdateModelMixin):
    retrievers = [UUIDIdentifierModelRetriever, CodeIdentifierModelRetriever]
    serializer_class = ClaimResponseSerializer
    permission_classes = (FHIRApiClaimPermissions,)

    def get_queryset(self):
        queryset = Claim.get_queryset(None, self.request.user).order_by('validity_from')
        return ValidityFromRequestParameterFilter(self.request).filter_queryset(queryset)



from rest_framework import status
from rest_framework.test import APITestCase
from api_fhir_r4.tests import GenericFhirAPITestMixin
from api_fhir_r4.configurations import GeneralConfiguration


class CodeSystemPHOrganizationLActivityAPITests(GenericFhirAPITestMixin, APITestCase):
    base_url = GeneralConfiguration.get_base_url() + 'CodeSystem/organization-ph-activity/'

    def setUp(self):
        self._EXPECTED_COUNT = 5

    def test_get_bad_authorization(self):
        response = self.client.get(self.base_url, data=None, format='json')
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_get_should_return_code_system(self):
        self.login()
        response = self.client.get(self.base_url, data=None, format='json')
        response_data = response.json()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response_data['count'], self._EXPECTED_COUNT)
        self.assertEqual(response_data['name'], 'OrganizationPHActivityCS')
        self.assertEqual(response_data['title'], 'Activity (Organization)')




from typing import Union, List, Tuple, Iterable

from django.core.exceptions import ValidationError
from django.core.validators import URLValidator

import core.datetimes.ad_datetime
from api_fhir_r4.converters import BaseFHIRConverter, ReferenceConverterMixin
from api_fhir_r4.models import Subscription, SubscriptionNotificationResult
from api_fhir_r4.subscriptions.notificationClient import RestSubscriptionNotificationClient, \
    SubscriberNotificationOutput
from core.models import HistoryModel, VersionedModel


class RestSubscriptionNotificationManager:

    def __init__(self,  fhir_converter: BaseFHIRConverter,
                 client: RestSubscriptionNotificationClient = None):
        if client is None:
            client = RestSubscriptionNotificationClient()
        self.client = client
        self.fhir_converter = fhir_converter

    def notify_subscribers_with_resource(
            self, imis_resource: Union[HistoryModel, VersionedModel], subscribers: List[Subscription])\
            -> Iterable[SubscriptionNotificationResult]:
        fhir_content = self._resource_to_fhir(imis_resource)
        valid, rejected = self._validate_subscribers(subscribers)
        result = self.client.propagate_notifications(fhir_content, valid)
        combined_result = [*result, *rejected]
        return self._handle_notification_results(combined_result)

    def _validate_subscribers(self, subscribers: List[Subscription]) \
            -> Tuple[List[Subscription], List[SubscriberNotificationOutput]]:
        url_validator = URLValidator()
        valid_subscribers = []
        rejected = []
        for subscriber in subscribers:
            try:
                url_validator(subscriber.endpoint)
                if subscriber.channel != Subscription.SubscriptionChannel.REST_HOOK:
                    raise ValidationError(f'Subscriber {subscriber} not eligible for REST notification. '
                                          f'Subscriber eligible for channel {subscriber.channel}.')
                valid_subscribers.append(subscriber)

            except ValidationError as e:
                rejected.append(
                    SubscriberNotificationOutput(
                        subscriber, False, F'Validation not passed, reason: {e}')
                )
        return valid_subscribers, rejected

    def _resource_to_fhir(self, imis_resource: Union[HistoryModel, VersionedModel]) -> dict:
        return self.fhir_converter.to_fhir_obj(imis_resource, ReferenceConverterMixin.UUID_REFERENCE_TYPE).dict()

    def _handle_notification_results(self, notification_result: Iterable[SubscriberNotificationOutput])\
            -> Iterable[SubscriptionNotificationResult]:
        return [self.__save_result_in_db(result) for result in notification_result]

    def __save_result_in_db(self, result: SubscriberNotificationOutput):
        new_entry = SubscriptionNotificationResult(
            subscription=result.subscription,
            error=str(result.reason_of_failure) if result.reason_of_failure else None,
            notified_successfully=result.notification_success,
            notification_time=core.datetimes.ad_datetime.AdDatetime.now()
        )
        new_entry.save()
        return new_entry




from api_fhir_r4.converters import InvoiceConverter
from api_fhir_r4.serializers import BaseFHIRSerializer


class InvoiceSerializer(BaseFHIRSerializer):
    fhirConverter = InvoiceConverter()

    def create(self, validated_data):
        pass

    def update(self, instance, validated_data):
        pass



from django.utils.translation import gettext as _
from rest_framework.test import APITestCase
from fhir.resources.insuranceplan import InsurancePlan
from api_fhir_r4.tests import GenericFhirAPITestMixin, FhirApiCreateTestMixin, \
    FhirApiUpdateTestMixin, FhirApiReadTestMixin
from api_fhir_r4.configurations import  GeneralConfiguration


class InsurancePlanAPITests(GenericFhirAPITestMixin, FhirApiReadTestMixin, APITestCase):

    base_url = GeneralConfiguration.get_base_url()+'InsurancePlan/'
    _test_json_path = "/test/test_insurance_plan.json"
    _TEST_PRODUCT_CODE = "Test0001"
    _TEST_MAX_INSTALLMENTS = 4

    def setUp(self):
        super(InsurancePlanAPITests, self).setUp()

    def verify_updated_obj(self, updated_obj):
        self.assertTrue(isinstance(updated_obj, InsurancePlan))
        max_installments_data = None
        for extension in updated_obj.extension:
            if "max-installments" in extension.url:
                max_installments_data = extension
        self.assertEqual(self._TEST_MAX_INSTALLMENTS, max_installments_data.valueUnsignedInt)

    def update_resource(self, data):
        for extension in data["extension"]:
            if "max-installments" in extension["url"]:
                extension["valueUnsignedInt"] = self._TEST_MAX_INSTALLMENTS

    def update_payload_missing_code_identifier(self, data):
        for i in range(len(data["identifier"])):
            if data["identifier"][i]["type"]["coding"][0]["code"] == "IC":
                del data["identifier"][i]
                return data



from api_fhir_r4.converters import ClaimResponseConverter
from api_fhir_r4.models import ClaimResponseV2 as ClaimResponse
from api_fhir_r4.tests import ClaimResponseTestMixin
from api_fhir_r4.tests.mixin import ConvertToFhirTestMixin, ConvertJsonToFhirTestMixin


class ClaimResponseConverterTestCase(ClaimResponseTestMixin,
                                     ConvertToFhirTestMixin,
                                     ConvertJsonToFhirTestMixin):
    converter = ClaimResponseConverter
    fhir_resource = ClaimResponse
    json_repr = 'test/test_claimResponse.json'



from api_fhir_r4.converters.claimConverter import ClaimConverter
from api_fhir_r4.models import ClaimV2 as Claim
from api_fhir_r4.tests import ClaimTestMixin
from api_fhir_r4.tests.mixin import ConvertToImisTestMixin, ConvertToFhirTestMixin, ConvertJsonToFhirTestMixin


class ClaimConverterTestCase(ClaimTestMixin,
                             ConvertToImisTestMixin,
                             ConvertToFhirTestMixin,
                             ConvertJsonToFhirTestMixin):
    converter = ClaimConverter
    fhir_resource = Claim
    json_repr = 'test/test_claim.json'



import json
import os

from rest_framework import status
from rest_framework.test import APITestCase

from api_fhir_r4.configurations import GeneralConfiguration
from api_fhir_r4.tests import GenericFhirAPITestMixin, FhirApiCreateTestMixin
from api_fhir_r4.tests.mixin.logInMixin import LogInMixin


class LoginAPITests(GenericFhirAPITestMixin, FhirApiCreateTestMixin, APITestCase, LogInMixin):
    base_url = GeneralConfiguration.get_base_url() + 'login/'
    _test_json_path = "/test/test_login.json"
    _test_json_path_wrong_credentials = "/tests/test/test_login_bad_credentials.json"
    _test_json_path_wrong_payload = "/tests/test/test_login_bad_payload.json"
    _TEST_EXPECTED_NAME = "UPDATED_NAME"
    _test_request_data_wrong_credentials = None
    _test_request_data_bad_payload = None

    def setUp(self):
        super(LoginAPITests, self).setUp()
        dir_path = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
        json_representation = open(dir_path + self._test_json_path_wrong_credentials).read()
        self._test_request_data_wrong_credentials = json.loads(json_representation)
        json_representation = open(dir_path + self._test_json_path_wrong_payload).read()
        self._test_request_data_bad_payload = json.loads(json_representation)

    def get_bundle_from_json_response(self, response):
        pass

    def test_post_should_create_correctly(self):
        self.get_or_create_user_api()
        response = self.client.post(self.base_url, data=self._test_request_data, format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_post_should_give_unathorized(self):
        response = self.client.post(self.base_url, data=self._test_request_data_wrong_credentials, format='json')
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_post_should_give_bad_request(self):
        response = self.client.post(self.base_url, data=self._test_request_data_bad_payload, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_get_should_required_login(self):
        pass



from fhir.resources.fhirtypes import ReferenceType
from typing import Iterable, List

from fhir.resources.reference import Reference

from api_fhir_r4.apps import logger
from api_fhir_r4.converters import BaseFHIRConverter
from api_fhir_r4.exceptions import FHIRRequestProcessException
from django.utils.translation import gettext as _


class ContainedResourceProcessException(FHIRRequestProcessException):
    pass


def get_converted_contained_resource(
        contained: Iterable,
        reference: ReferenceType,
        converter_for_contained: BaseFHIRConverter,
        audit_user_id: int
):
    def validate_reference(ref):
        if isinstance(ref, str) and '/' in ref and ref[0] == '#':
            return True
        else:
            raise ContainedResourceProcessException(
                'Invalid reference.reference format or type, '
                'should be string "#{reference_type}/{reference_id}"'
            )

    def _validate_matching_contained(matching: List):
        if len(matching) > 1:
            logger.warning([
                f'More than one contained resource definition found for reference '
                f'{reference}, references should be unique.']
            )
            logger.debug(f"Dupliacted contained resources: {matching}")

    if reference:
        reference = reference.reference
        validate_reference(reference)
        path, resource_id = reference.split('/', maxsplit=1)
        resource_type = path[1:]
        contained = [
            n for n in contained if n.resource_type == resource_type and n.id == resource_id
        ]

        _validate_matching_contained(contained)
        if len(contained) > 0:
            return converter_for_contained.to_imis_obj(contained[0].__dict__, audit_user_id)

    return None


def get_from_contained_or_by_reference(fhir_reference, contained, converter, audit_user_id):
    """
    For given fhir reference, function checks if definition is provided in contained resource. If contained resource
    exists it's converted to IMIS type using converter and returned.
    If definition is not found in contained resource then internal value for reference is returned.

    @param fhir_reference: FHIR Reference to IMIS resource
    @param contained: Collection of contained resources
    @param converter: BaseFHIRConverter, has to implement get_imis_obj_by_fhir_reference and to_imis_obj
    @param audit_user_id: ID of user performing operation
    """

    value = None
    if fhir_reference:
        if contained:
            value = get_converted_contained_resource(contained, fhir_reference, converter, audit_user_id)
        value = value or converter.get_imis_obj_by_fhir_reference(fhir_reference)
    return value



from rest_framework import viewsets

from api_fhir_r4.mixins import MultiIdentifierRetrieverMixin, MultiIdentifierUpdateMixin
from api_fhir_r4.model_retrievers import CodeIdentifierModelRetriever, UUIDIdentifierModelRetriever
from api_fhir_r4.permissions import FHIRApiMedicationPermissions
from api_fhir_r4.serializers import MedicationSerializer
from api_fhir_r4.views.fhir.base import BaseFHIRView
from api_fhir_r4.views.filters import ValidityFromRequestParameterFilter
from medical.models import Item


class MedicationViewSet(BaseFHIRView, MultiIdentifierRetrieverMixin, MultiIdentifierUpdateMixin, viewsets.ModelViewSet):
    retrievers = [UUIDIdentifierModelRetriever, CodeIdentifierModelRetriever]
    serializer_class = MedicationSerializer
    permission_classes = (FHIRApiMedicationPermissions,)

    def list(self, request, *args, **kwargs):
        queryset = self.get_queryset()
        identifier = request.GET.get("identifier")
        if identifier:
            return self.retrieve(request, *args, **{**kwargs, 'identifier': identifier})
        else:
            queryset = queryset.filter(validity_to__isnull=True)
        serializer = MedicationSerializer(self.paginate_queryset(queryset), many=True)
        return self.get_paginated_response(serializer.data)

    def retrieve(self, *args, **kwargs):
        response = super().retrieve(self, *args, **kwargs)
        return response

    def get_queryset(self):
        queryset = Item.get_queryset(None, self.request.user).order_by('validity_from')
        return ValidityFromRequestParameterFilter(self.request).filter_queryset(queryset)



import base64
import hashlib
import re
from urllib.parse import urljoin

from typing import Type

from claim.services import ClaimElementSubmit
from claim.apps import ClaimConfig
from claim.models import Claim, ClaimItem, ClaimService, ClaimAttachment

from api_fhir_r4.containedResources.converterUtils import get_from_contained_or_by_reference
from api_fhir_r4.mapping.claimMapping import ClaimPriorityMapping, ClaimVisitTypeMapping
from medical.models import Diagnosis
from django.utils.translation import gettext as _

from api_fhir_r4.configurations import R4IdentifierConfig, R4ClaimConfig, GeneralConfiguration
from api_fhir_r4.converters import BaseFHIRConverter, ReferenceConverterMixin, MedicationConverter, \
    ActivityDefinitionConverter
from api_fhir_r4.converters.patientConverter import PatientConverter
from api_fhir_r4.converters.healthFacilityOrganisationConverter import HealthFacilityOrganisationConverter
from api_fhir_r4.converters.claimAdminPractitionerConverter import ClaimAdminPractitionerConverter
from api_fhir_r4.models import ClaimV2 as FHIRClaim, ClaimInsuranceV2 as ClaimInsurance
from fhir.resources.attachment import Attachment
from fhir.resources.period import Period
from fhir.resources.claim import ClaimDiagnosis, ClaimSupportingInfo, ClaimItem as FHIRClaimItem

from api_fhir_r4.utils import TimeUtils, FhirUtils, DbManagerUtils


class ClaimConverter(BaseFHIRConverter, ReferenceConverterMixin):

    @classmethod
    def to_fhir_obj(cls, imis_claim, reference_type=ReferenceConverterMixin.UUID_REFERENCE_TYPE):
        fhir_claim = cls.build_fhir_obj_with_required_fields(imis_claim)
        cls.build_fhir_identifiers(fhir_claim, imis_claim)
        cls.build_fhir_pk(fhir_claim, imis_claim, reference_type)
        cls.build_fhir_provider(fhir_claim, imis_claim, reference_type)
        cls.build_fhir_patient(fhir_claim, imis_claim, reference_type)
        cls.build_fhir_enterer(fhir_claim, imis_claim, reference_type)
        cls.build_fhir_type(fhir_claim, imis_claim)
        cls.build_fhir_priority(fhir_claim)
        cls.build_fhir_insurance(fhir_claim, imis_claim, reference_type)
        cls.build_fhir_billable_period(fhir_claim, imis_claim)
        cls.build_fhir_diagnoses(fhir_claim, imis_claim)
        cls.build_fhir_total(fhir_claim, imis_claim)
        cls.build_fhir_items(fhir_claim, imis_claim, reference_type)
        cls.build_fhir_supporting_info(fhir_claim, imis_claim)
        cls.build_fhir_attachments(fhir_claim, imis_claim)
        return fhir_claim

    @classmethod
    def to_imis_obj(cls, fhir_claim, audit_user_id):
        errors = []
        fhir_claim = FHIRClaim(**fhir_claim)
        imis_claim = Claim()
        imis_claim.audit_user_id = audit_user_id
        cls.build_imis_date_claimed(imis_claim, fhir_claim, errors)
        cls.build_imis_health_facility(errors, fhir_claim, imis_claim, audit_user_id=audit_user_id)
        cls.build_imis_identifier(imis_claim, fhir_claim, errors)
        cls.build_imis_patient(imis_claim, fhir_claim, errors, audit_user_id=audit_user_id)
        cls.build_imis_date_range(imis_claim, fhir_claim, errors)
        cls.build_imis_diagnoses(imis_claim, fhir_claim, errors)
        cls.build_imis_total_claimed(imis_claim, fhir_claim, errors)
        cls.build_imis_claim_admin(imis_claim, fhir_claim, errors, audit_user_id=audit_user_id)
        cls.build_imis_visit_type(imis_claim, fhir_claim, errors)
        cls.build_imis_supporting_info(imis_claim, fhir_claim, errors)
        cls.build_imis_submit_items_and_services(imis_claim, fhir_claim, errors, audit_user_id)
        cls.check_errors(errors)
        return imis_claim

    @classmethod
    def get_fhir_code_identifier_type(cls):
        return R4IdentifierConfig.get_fhir_claim_code_type()

    @classmethod
    def get_reference_obj_id(cls, imis_claim):
        return imis_claim.id

    @classmethod
    def get_reference_obj_uuid(cls, imis_claim):
        return imis_claim.uuid

    @classmethod
    def get_reference_obj_code(cls, imis_claim):
        return imis_claim.code

    @classmethod
    def get_fhir_resource_type(cls):
        return FHIRClaim

    @classmethod
    def get_imis_obj_by_fhir_reference(cls, reference, errors=None):
        imis_claim_code = cls.get_resource_id_from_reference(reference)
        return DbManagerUtils.get_object_or_none(Claim, code=imis_claim_code)

    @classmethod
    def build_imis_date_claimed(cls, imis_claim, fhir_claim, errors):
        if fhir_claim.created:
            imis_claim.date_claimed = TimeUtils.str_to_date(fhir_claim.created)
        cls.valid_condition(not imis_claim.date_claimed, _('Missing `created` attribute'), errors)

    @classmethod
    def build_fhir_identifiers(cls, fhir_claim, imis_claim):
        identifiers = []
        cls.build_all_identifiers(identifiers, imis_claim)
        fhir_claim.identifier = identifiers

    @classmethod
    def build_imis_identifier(cls, imis_claim, fhir_claim, errors):
        value = cls.get_fhir_identifier_by_code(fhir_claim.identifier, cls.get_fhir_code_identifier_type())
        if value:
            imis_claim.code = value
        cls.valid_condition(not imis_claim.code, _('Missing or invalid `identifier` attribute'), errors)

    @classmethod
    def build_fhir_obj_with_required_fields(cls, imis_claim):
        fhir_claim_dict = {}
        cls.build_fhir_status(fhir_claim_dict)
        cls.build_fhir_use(fhir_claim_dict)
        cls.build_fhir_created(fhir_claim_dict, imis_claim)
        return FHIRClaim(**fhir_claim_dict)

    @classmethod
    def build_imis_patient(cls, imis_claim, fhir_claim, errors, audit_user_id):
        insuree = get_from_contained_or_by_reference(
            fhir_claim.patient, fhir_claim.contained, PatientConverter, audit_user_id)

        if insuree:
            imis_claim.insuree = insuree
            imis_claim.insuree_chf_code = insuree.chf_id
        cls.valid_condition(not imis_claim.insuree, _('Missing or invalid `patient` reference'), errors)

    @classmethod
    def build_imis_health_facility(cls, errors, fhir_claim, imis_claim, audit_user_id):
        health_facility = get_from_contained_or_by_reference(
            fhir_claim.provider, fhir_claim.contained, HealthFacilityOrganisationConverter, audit_user_id
        )
        if health_facility:
            imis_claim.health_facility = health_facility
            imis_claim.health_facility_code = health_facility.code
        cls.valid_condition(not imis_claim.health_facility, _('Missing or invalid `provider` reference'), errors)

    @classmethod
    def build_fhir_billable_period(cls, fhir_claim, imis_claim):
        billable_period = Period.construct()
        billable_period.start = imis_claim.date_from.isoformat()
        if imis_claim.date_to:
            billable_period.end = imis_claim.date_to.isoformat()
        else:
            billable_period.end = billable_period.start
        fhir_claim.billablePeriod = billable_period

    @classmethod
    def build_imis_date_range(cls, imis_claim, fhir_claim, errors):
        billable_period = fhir_claim.billablePeriod
        if billable_period and hasattr(billable_period, "start"):
            if billable_period.start:
                imis_claim.date_from = TimeUtils.str_to_date(billable_period.start)
            if billable_period.end:
                imis_claim.date_to = TimeUtils.str_to_date(billable_period.end)
        cls.valid_condition(not imis_claim.date_from, _("Missing or invalid 'billable_period' attribute"), errors)

    @classmethod
    def build_fhir_diagnoses(cls, fhir_claim, imis_claim):
        fhir_diagnoses = []
        imis_diagnoses = [imis_claim.icd, imis_claim.icd_1, imis_claim.icd_2, imis_claim.icd_3, imis_claim.icd_4]
        for icd in imis_diagnoses:
            if icd:
                cls.build_fhir_diagnosis(fhir_diagnoses, icd)
        fhir_claim.diagnosis = fhir_diagnoses

    @classmethod
    def build_fhir_diagnosis(cls, diagnoses, icd):
        base = GeneralConfiguration.get_system_base_url()
        system = urljoin(base, R4ClaimConfig.get_fhir_claim_diagnosis_system())
        diagnosis_codeable_concept = cls.build_codeable_concept(icd.code, system=system, display=icd.name)
        claim_diagnosis_data = {'sequence': FhirUtils.get_next_array_sequential_id(diagnoses),
                                'diagnosisCodeableConcept': diagnosis_codeable_concept.dict()}
        diagnoses.append(ClaimDiagnosis(**claim_diagnosis_data))

    @classmethod
    def build_imis_diagnoses(cls, imis_claim, fhir_claim, errors):
        diagnoses = fhir_claim.diagnosis
        icd_fields_suffixes = [["", "_code"], ["_1", "1_code"], ["_2", "2_code"], ["_3", "3_code"], ["_4", "4_code"]]
        if diagnoses:
            for diagnosis, field_suffix in zip(diagnoses, icd_fields_suffixes):
                diagnosis_code = cls.get_imis_diagnosis_code(diagnosis)
                diagnosis_obj = cls.get_imis_diagnosis_by_code(diagnosis_code)
                setattr(imis_claim, "icd" + field_suffix[0], diagnosis_obj)
                setattr(imis_claim, "icd" + field_suffix[1], diagnosis_obj.code)
        cls.valid_condition(imis_claim.icd is None, _('Missing or invalid `diagnosis` attribute'), errors)

    @classmethod
    def get_imis_diagnosis_by_code(cls, icd_code):
        return Diagnosis.objects.get(code=icd_code)

    @classmethod
    def get_imis_diagnosis_code(cls, diagnosis):
        coding = cls.get_first_coding_from_codeable_concept(diagnosis.diagnosisCodeableConcept)
        return coding.code

    @classmethod
    def build_fhir_total(cls, fhir_claim, imis_claim):
        fhir_claim.total = cls.build_fhir_money(imis_claim.claimed)

    @classmethod
    def build_imis_total_claimed(cls, imis_claim, fhir_claim, errors):
        total_money = fhir_claim.total
        if total_money is not None:
            imis_claim.claimed = total_money.value
        cls.valid_condition(not imis_claim.claimed, _('Missing `total` attribute'), errors)

    @classmethod
    def build_imis_claim_admin(cls, imis_claim, fhir_claim, errors, audit_user_id):
        admin = get_from_contained_or_by_reference(
            fhir_claim.enterer, fhir_claim.contained, ClaimAdminPractitionerConverter, audit_user_id
        )

        if admin:
            imis_claim.admin = admin
            imis_claim.claim_admin_code = admin.code
        cls.valid_condition(imis_claim.admin is None, _('Missing or invalid `enterer` reference'), errors)

    @classmethod
    def build_fhir_type(cls, fhir_claim, imis_claim):
        mapping = ClaimVisitTypeMapping.fhir_claim_visit_type_coding[imis_claim.visit_type]
        fhir_claim.type = cls.build_codeable_concept_from_coding(cls.build_fhir_mapped_coding(mapping))

    @classmethod
    def build_imis_visit_type(cls, imis_claim, fhir_claim, errors):
        if fhir_claim.type:
            claim_type = cls.get_first_coding_from_codeable_concept(fhir_claim.type)
            if hasattr(claim_type, "code"):
                imis_claim.visit_type = claim_type.code
        cls.valid_condition(not imis_claim.visit_type, _('Missing or invalid `type` attribute'), errors)

    @classmethod
    def build_fhir_supporting_info(cls, fhir_claim, imis_claim):
        supporting_info = []
        guarantee_id_code = R4ClaimConfig.get_fhir_claim_information_guarantee_id_code()
        cls.build_fhir_string_information(supporting_info, guarantee_id_code, imis_claim.guarantee_id)
        explanation_code = R4ClaimConfig.get_fhir_claim_information_explanation_code()
        cls.build_fhir_string_information(supporting_info, explanation_code, imis_claim.explanation)
        fhir_claim.supportingInfo = supporting_info

    @classmethod
    def build_imis_supporting_info(cls, imis_claim, fhir_claim, errors):
        if not hasattr(imis_claim, "claim_attachments") or imis_claim.claim_attachments is not list:
            imis_claim.claim_attachments = []
        if fhir_claim.supportingInfo:
            for supporting_info in fhir_claim.supportingInfo:
                category_coding = cls.get_first_coding_from_codeable_concept(supporting_info.category)
                category = category_coding.code
                if category == R4ClaimConfig.get_fhir_claim_information_guarantee_id_code():
                    imis_claim.guarantee_id = supporting_info.valueString
                elif category == R4ClaimConfig.get_fhir_claim_information_explanation_code():
                    imis_claim.explanation = supporting_info.valueString
                elif category == R4ClaimConfig.get_fhir_claim_attachment_code():
                    claim_attachment = cls.build_attachment_from_value(supporting_info.valueAttachment)
                    imis_claim.claim_attachments.append(claim_attachment)
                else:
                    cls.valid_condition(True, _('Unknown supporting info category: `%s`') % category, errors)

    @classmethod
    def build_fhir_string_information(cls, supporting_info, code, value_string):
        if value_string:
            supporting_info_entry = ClaimSupportingInfo.construct()
            supporting_info_entry.sequence = FhirUtils.get_next_array_sequential_id(supporting_info)
            base = GeneralConfiguration.get_system_base_url()
            system = urljoin(base, R4ClaimConfig.get_fhir_claim_supporting_info_system())
            category = cls.build_codeable_concept(code, system)
            supporting_info_entry.category = category
            supporting_info_entry.valueString = value_string
            supporting_info.append(supporting_info_entry)

    @classmethod
    def build_fhir_items(cls, fhir_claim, imis_claim, reference_type):
        fhir_claim.item = []
        cls.build_fhir_items_for_imis_items(fhir_claim, imis_claim, reference_type)
        cls.build_fhir_items_for_imis_services(fhir_claim, imis_claim, reference_type)

    @classmethod
    def build_fhir_items_for_imis_items(cls, fhir_claim, imis_claim, reference_type):
        for claim_item in imis_claim.items.filter(validity_to=None).all():
            if claim_item:
                item_type = R4ClaimConfig.get_fhir_claim_item_code()
                cls.build_fhir_item(fhir_claim, claim_item.item.code, item_type, claim_item, reference_type)

    @classmethod
    def get_imis_items_for_claim(cls, imis_claim):
        items = []
        if imis_claim and imis_claim.id:
            items = ClaimItem.objects.filter(claim_id=imis_claim.id)
        return items

    @classmethod
    def build_fhir_item(cls, fhir_claim, code, item_type, claim_item, reference_type):
        fhir_item = FHIRClaimItem.construct()
        fhir_item.sequence = FhirUtils.get_next_array_sequential_id(fhir_claim.item)
        fhir_item.unitPrice = cls.build_fhir_money(claim_item.price_asked)
        fhir_item.quantity = cls.build_fhir_quantity(claim_item.qty_provided)
        fhir_item.productOrService = cls.build_simple_codeable_concept(code)
        fhir_item.category = cls.build_simple_codeable_concept(item_type)
        fhir_item.extension = []

        if item_type == R4ClaimConfig.get_fhir_claim_item_code():
            medication = cls.build_medication_extension(claim_item, reference_type)
            fhir_item.extension.append(medication)

        elif item_type == R4ClaimConfig.get_fhir_claim_service_code():
            activity_definition = cls.build_activity_definition_extension(claim_item, reference_type)
            fhir_item.extension.append(activity_definition)

        fhir_claim.item.append(fhir_item)

    @classmethod
    def build_fhir_items_for_imis_services(cls, fhir_claim, imis_claim, reference_type):
        for claim_service in imis_claim.services.filter(validity_to=None).all():
            if claim_service:
                item_type = R4ClaimConfig.get_fhir_claim_service_code()
                cls.build_fhir_item(fhir_claim, claim_service.service.code, item_type, claim_service, reference_type)

    @classmethod
    def get_imis_services_for_claim(cls, imis_claim):
        services = []
        if imis_claim and imis_claim.id:
            services = ClaimService.objects.filter(claim_id=imis_claim.id)
        return services

    @classmethod
    def build_medication_extension(cls, item, reference_type):
        base = GeneralConfiguration.get_system_base_url()
        url = urljoin(base, R4ClaimConfig.get_fhir_item_reference_extension_system())
        reference = cls.build_fhir_resource_reference(item.item, type='Medication', reference_type=reference_type)
        return cls.build_fhir_reference_extension(reference, url)

    @classmethod
    def build_activity_definition_extension(cls, service, reference_type):
        base = GeneralConfiguration.get_system_base_url()
        url = urljoin(base, R4ClaimConfig.get_fhir_item_reference_extension_system())
        reference = cls.build_fhir_resource_reference(service.service, type='ActivityDefinition',
                                                      reference_type=reference_type)
        return cls.build_fhir_reference_extension(reference, url)

    @classmethod
    def build_imis_submit_items_and_services(cls, imis_claim, fhir_claim, errors, audit_user_id):
        imis_items = []
        imis_services = []
        if fhir_claim.item:
            for item in fhir_claim.item:
                category = item.category.text
                if category == R4ClaimConfig.get_fhir_claim_item_code():
                    cls.build_imis_submit_item(imis_items, item, fhir_claim, audit_user_id)
                elif category == R4ClaimConfig.get_fhir_claim_service_code():
                    cls.build_imis_submit_service(imis_services, item, fhir_claim, audit_user_id)
                else:
                    cls.valid_condition(True, _('Unknown item category: `%s`') % category, errors)

        # added additional attributes which will be used to create ClaimRequest in serializer
        imis_claim.submit_items = imis_items
        imis_claim.submit_services = imis_services

    @classmethod
    def build_imis_submit_item(cls, imis_items, fhir_item, fhir_claim, audit_user_id):
        item = cls.__get_provision_from_contained_or_reference(
                fhir_item, fhir_claim,
                MedicationConverter,
                audit_user_id)
        claim_item = ClaimItem(
            qty_provided=cls.get_fhir_item_qty_provided(fhir_item),
            price_asked=cls.get_fhir_item_price_asked(fhir_item),
            item=item
        )
        claim_item.code = item.code
        imis_items.append(claim_item)

    @classmethod
    def build_imis_submit_service(cls, imis_services, fhir_item, fhir_claim, audit_user_id):
        claim_service = ClaimService(
            qty_provided=cls.get_fhir_item_qty_provided(fhir_item),
            price_asked=cls.get_fhir_item_price_asked(fhir_item),
            service=cls.__get_provision_from_contained_or_reference(
                fhir_item, fhir_claim,
                ActivityDefinitionConverter,
                audit_user_id)
        )
        claim_service.code = claim_service.service.code
        imis_services.append(claim_service)

    @classmethod
    def __get_provision_from_contained_or_reference(cls, fhir_provision, fhir_claim, converter, audit_user_id):
        # Items and services are referenced by extension with ValueReference
        reference = fhir_provision.extension[0].valueReference
        return get_from_contained_or_by_reference(
            reference, fhir_claim.contained, converter, audit_user_id)

    @classmethod
    def _get_imis_claim_provision(cls, fhir_item, submission_type: Type[ClaimElementSubmit]):
        price_asked = cls.get_fhir_item_price_asked(fhir_item)
        quantity = cls.get_fhir_item_qty_provided(fhir_item)
        code = cls.get_fhir_item_code(fhir_item)
        return submission_type(code, quantity, price_asked).to_claim_provision()

    @classmethod
    def get_fhir_item_code(cls, fhir_item):
        item_code = None
        if fhir_item.productOrService:
            item_code = fhir_item.productOrService.text
        return item_code

    @classmethod
    def get_fhir_item_qty_provided(cls, fhir_item):
        qty_provided = None
        if fhir_item.quantity:
            qty_provided = fhir_item.quantity.value
        return qty_provided

    @classmethod
    def get_fhir_item_price_asked(cls, fhir_item):
        price_asked = None
        if fhir_item.unitPrice:
            price_asked = fhir_item.unitPrice.value
        return price_asked

    @classmethod
    def build_fhir_use(cls, fhir_claim_dict):
        fhir_claim_dict['use'] = "claim"

    @classmethod
    def build_fhir_priority(cls, fhir_claim):
        mapping = ClaimPriorityMapping.fhir_priority_coding['normal']
        fhir_claim.priority = cls.build_codeable_concept_from_coding(cls.build_fhir_mapped_coding(mapping))

    @classmethod
    def build_fhir_status(cls, fhir_claim_dict):
        fhir_claim_dict['status'] = "active"

    @classmethod
    def build_fhir_created(cls, fhir_claim_dict, imis_claim):
        fhir_claim_dict['created'] = imis_claim.date_claimed.isoformat()

    @classmethod
    def build_fhir_insurance(cls, fhir_claim, imis_claim, reference_type):
        policies = list(imis_claim.insuree.insuree_policies.all())
        if policies:
            sorted_by_enrol = sorted(list(policies), key=lambda x: x.enrollment_date, reverse=True)
            latest = sorted_by_enrol[0]
            claim_insurance_data = {'focal': True, 'sequence': 1}
            insurance = ClaimInsurance(**claim_insurance_data)
            insurance.coverage = cls.build_fhir_resource_reference(latest.policy, type="Coverage",
                                                                   reference_type=reference_type)
            fhir_claim.insurance = [insurance]

    @classmethod
    def build_fhir_attachments(cls, fhir_claim, imis_claim):
        attachments = ClaimAttachment.objects.filter(claim=imis_claim)

        if not fhir_claim.supportingInfo:
            fhir_claim.supportingInfo = []

        for attachment in attachments:
            supporting_info_element = cls.build_attachment_supporting_info_element(attachment)
            if fhir_claim.supportingInfo is not list:
                fhir_claim.supportingInfo = [supporting_info_element]
            else:
                fhir_claim.supportingInfo.append(supporting_info_element)

    @classmethod
    def build_attachment_supporting_info_element(cls, imis_attachment):
        supporting_info_element = ClaimSupportingInfo.construct()

        supporting_info_element.category = cls.build_attachment_supporting_info_category()
        supporting_info_element.valueAttachment = cls.build_fhir_value_attachment(imis_attachment)
        return supporting_info_element

    @classmethod
    def build_attachment_supporting_info_category(cls):
        category_code = R4ClaimConfig.get_fhir_claim_attachment_code()
        system = R4ClaimConfig.get_fhir_claim_attachment_system()
        category = cls.build_codeable_concept(category_code, system, category_code)
        category.coding[0].display = category_code.capitalize()
        return category

    @classmethod
    def build_fhir_value_attachment(cls, imis_attachment):
        attachment = Attachment.construct()
        attachment.creation = imis_attachment.date.isoformat()
        attachment.data = cls.get_attachment_content(imis_attachment)
        attachment.contentType = imis_attachment.mime
        attachment.title = imis_attachment.filename
        return attachment

    @classmethod
    def get_attachment_content(cls, imis_attachment):
        file_root = ClaimConfig.claim_attachments_root_path

        if file_root and imis_attachment.url:
            with open('%s/%s' % (ClaimConfig.claim_attachments_root_path, imis_attachment.url), "rb") as file:
                return base64.b64encode(file.read())
        elif not imis_attachment.url and imis_attachment.document:
            return imis_attachment.document
        else:
            return None

    @classmethod
    def build_attachment_from_value(cls, valueAttachment: Attachment):
        allowed_mime_regex = R4ClaimConfig.get_allowed_fhir_claim_attachment_mime_types_regex()
        mime_validation = re.compile(allowed_mime_regex, re.IGNORECASE)

        if not mime_validation.match(valueAttachment.contentType):
            raise ValueError(F'Mime type {valueAttachment.contentType} not allowed')

        if valueAttachment.hash:
            cls.validateHash(valueAttachment.hash, valueAttachment.data)

        attachment_data = {
            'title': valueAttachment.title,
            'filename': valueAttachment.title,
            'document': valueAttachment.data,
            'mime': valueAttachment.contentType,
            'date': TimeUtils.str_to_date(valueAttachment.creation)
        }
        return attachment_data

    @classmethod
    def validateHash(cls, expected_hash, data):
        actual_hash = hashlib.sha1(data.encode('utf-8')).hexdigest()
        if actual_hash.casefold() != expected_hash.casefold():
            raise ValueError('Hash for data file is incorrect')

    @classmethod
    def build_fhir_provider(cls, fhir_claim, imis_claim, reference_type):
        fhir_claim.provider = cls.build_fhir_resource_reference(imis_claim.health_facility,
                                                                type='Organization',
                                                                display=imis_claim.health_facility.code,
                                                                reference_type=reference_type)

    @classmethod
    def build_fhir_patient(cls, fhir_claim, imis_claim, reference_type):
        fhir_claim.patient = cls.build_fhir_resource_reference(imis_claim.insuree,
                                                               type='Patient',
                                                               display=imis_claim.insuree.chf_id,
                                                               reference_type=reference_type)

    @classmethod
    def build_fhir_enterer(cls, fhir_claim, imis_claim, reference_type):
        fhir_claim.enterer = cls.build_fhir_resource_reference(imis_claim.admin,
                                                               type='Practitioner',
                                                               display=imis_claim.admin.code,
                                                               reference_type=reference_type)



from drf_spectacular.authentication import SessionScheme, TokenScheme
from drf_spectacular.plumbing import build_bearer_security_scheme_object
from drf_spectacular.utils import inline_serializer
from rest_framework import fields


class CsrfExemptAuthenticationScheme(SessionScheme):
    target_class = 'api_fhir_r4.views.CsrfExemptSessionAuthentication'
    name = "Session Authentication (CSRF Exempt)"
    priority = 0
    exclude = True


class JWTAuthenticationScheme(TokenScheme):
    target_class = 'core.jwt_authentication.JWTAuthentication'
    name = "JWT Authentication"
    priority = 0

    def get_security_definition(self, auto_schema):
        return build_bearer_security_scheme_object(
            header_name='Authorization',
            token_prefix='Bearer',
        )


def get_inline_error_serializer():
    return inline_serializer(
        name='InlineErrorSerializer',
        fields={
            'resourceType': fields.CharField(),
            'issue': fields.ListField(child=fields.JSONField()),
        }
    )


def get_inline_login_request_serializer():
    return inline_serializer(
        name='InlineLoginRequestSerializer',
        fields={
            'username': fields.CharField(),
            'password': fields.CharField(),
        }
    )


def get_inline_login_200_response_serializer():
    return inline_serializer(
        name='InlineLoginResponseSerializer',
        fields={
            'token': fields.CharField(),
            'exp': fields.IntegerField(),
        }
    )



DEFAULT_CFG = {
    "default_audit_user_id": 1,
    "gender_codes": {
        "male": "M",
        "female": "F",
        "other": "O"
    },
    "base_url": "https://openimis.github.io/openimis_fhir_r4_ig/",
    "default_value_of_patient_head_attribute": False,
    "default_value_of_patient_card_issued_attribute": False,
    "default_value_of_location_offline_attribute": False,
    "default_value_of_location_care_type": "B",
    "default_response_page_size": 10,
    "claim_rule_engine_validation": True,
    "R4_fhir_identifier_type_config": {
        "system": "https://openimis.github.io/openimis_fhir_r4_ig/CodeSystem/openimis-identifiers",
        "fhir_code_for_imis_db_uuid_type": "UUID",
        "fhir_code_for_imis_db_id_type": "ACSN",
        "fhir_code_for_imis_generic_code": "Code",
        "fhir_code_for_imis_chfid_type": "SB",
        "fhir_code_for_imis_passport_type": "PPN",
        "fhir_code_for_imis_facility_id_type": "FI",
        "fhir_code_for_imis_claim_admin_code_type": "FILL",
        "fhir_code_for_imis_claim_code_type": "MR",
        "fhir_code_for_imis_location_code_type": "LC",
        "fhir_code_for_imis_diagnosis_code_type": "DC",
        "fhir_code_for_imis_item_code_type": "IC",
        "fhir_code_for_imis_service_code_type": "SC"
    },
    "R4_fhir_marital_status_config": {
        "system": "http://hl7.org/fhir/valueset-marital-status.html",
        "fhir_code_for_married": "M",
        "fhir_code_for_never_married": "S",
        "fhir_code_for_divorced": "D",
        "fhir_code_for_widowed": "W",
        "fhir_code_for_unknown": "U"
    },
    "R4_fhir_location_site_type": {
        "system": "http://hl7.org/fhir/v3/ServiceDeliveryLocationRoleType/vs.html",
        "fhir_code_for_hospital": "HOSP",
        "fhir_code_for_dispensary": "COMM",
        "fhir_code_for_health_center": "OF",
    },
    "R4_fhir_location_physical_type": {
        "system": "http://terminology.hl7.org/CodeSystem/location-physical-type.html",
        "fhir_code_for_region": "R",
        "fhir_code_for_district": "D",
        "fhir_code_for_ward": "W",
        "fhir_code_for_village": "V"
    },
    "R4_fhir_location_status_codes": {
        "fhir_code_for_active": "active",
        "fhir_code_for_inactive": "inactive",
    },
    "R4_fhir_hf_service_type": {
        "system": "http://hl7.org/fhir/valueset-service-type.html",
        "fhir_code_for_in_patient": "I",
        "fhir_code_for_out_patient": "O",
        "fhir_code_for_both": "B"
    },
    "R4_fhir_issue_type_config": {
        "fhir_code_for_exception": "exception",
        "fhir_code_for_not_found": "not-found",
        "fhir_code_for_informational": "informational"
    },
    "R4_fhir_claim_config": {
        "fhir_claim_information_guarantee_id_code": "guarantee",
        "fhir_claim_information_explanation_code": "explanation",
        "fhir_claim_item_explanation_code": "item_explanation",
        "fhir_claim_item_code": "item",
        "fhir_claim_service_code": "service",
        "fhir_claim_status_rejected_code": "rejected",
        "fhir_claim_status_entered_code": "entered",
        "fhir_claim_status_checked_code": "checked",
        "fhir_claim_status_processed_code": "processed",
        "fhir_claim_status_valuated_code": "valuated",
        "fhir_claim_item_status_code": "claim_item_status",
        "fhir_claim_item_status_passed_code": "passed",
        "fhir_claim_item_status_rejected_code": "rejected",
        "fhir_claim_item_general_adjudication_code": "general",
        "fhir_claim_item_rejected_reason_adjudication_code": "rejected_reason",
        "fhir_claim_organization_code": "openIMIS",
        "fhir_claim_attachment_code": "attachment",
        "fhir_claim_supporting_info_category_system": 'CodeSystem/claim-supporting-info-category',
        'fhir_claim_item_reference_extension_system': 'StructureDefinition/claim-item-reference',
        'fhir_claim_visit_type_system': 'CodeSystem/claim-visit-type',
        'fhir_claim_item_category_system': 'CodeSystem/claim-item-category',
        'fhir_claim_diagnosis_system': 'CodeSystem/diagnosis-ICD10-level1',
        "fhir_claim_allowed_mime_types_regex":
            '|'.join(['text\/.*', 'image\/png', 'image\/jpe?g', 'application\/msword', '.*doc.*'])
    },
    "R4_fhir_coverage_eligibility_config": {
        "fhir_serializer": "PolicyCoverageEligibilityRequestSerializer",
        "fhir_item_code": "item",
        "fhir_service_code": "service",
        "fhir_total_admissions_code": "total_admissions",
        "fhir_total_visits_code": "total_visits",
        "fhir_total_consultations_code": "total_consultations",
        "fhir_total_surgeries_code": "total_surgeries",
        "fhir_total_deliveries_code": "total_deliveries",
        "fhir_total_antenatal_code": "total_antenatal",
        "fhir_consultation_amount_code": "consultation_amount",
        "fhir_surgery_amount_code": "surgery_amount",
        "fhir_delivery_amount_code": "delivery_amount",
        "fhir_hospitalization_amount_code": "hospitalization_amount",
        "fhir_antenatal_amount_code": "antenatal_amount",
        "fhir_service_left_code": "service_left",
        "fhir_item_left_code": "item_left",
        "fhir_is_item_ok_code": "is_item_ok",
        "fhir_is_service_ok_code": "is_service_ok",
        "fhir_balance_code": "balance",
        "fhir_balance_default_category": "medical",
        "fhir_active_policy_status": ("A", 2)
    },
    "R4_fhir_communication_request_config": {
        "fhir_care_rendered_code": "CareRendered",
        "fhir_payment_asked_code": "PaymentAsked",
        "fhir_drug_prescribed_code": "DrugPrescribed",
        "fhir_drug_received_code": "DrugReceived",
        "fhir_asessment_code": "Asessment"
    },
    "R4_fhir_contract_config": {
        "fhir_contract_eo_signer_type": "EnrolmentOfficer",
        "fhir_contract_head_signer_type": "HeadOfFamily",
        "fhir_contract_insuree_role": "Insuree",
        "fhir_contract_dependant_role": "Dependant",
        "fhir_contract_executable_status": "Executable",
        "fhir_contract_renewed_status": "Renewed",
        "fhir_contract_policy_status": "Policy",
        "fhir_contract_Terminated_status": "Terminated"
    },
    "R4_fhir_coverage_config": {
        "fhir_family_refereence_code": "FamilyReference",
        "fhir_status_idle_code": "Idle",
        "fhir_status_active_code": "Active",
        "fhir_status_suspended_code": "Suspended",
        "fhir_status_expired_code": "Expired",
        "fhir_status_disputed_code": "Disputed",
        "fhir_item_code": "item",
        "fhir_service_code": "service",
        "fhir_practitioner_role_code": "Practitioner",
        "fhir_product_code": "Product",
        "fhir_effective_date_code": "EffectiveDate",
        "fhir_enroll_date_code": "EnrollDate"
    },
    "R4_fhir_organisation_config": {
        "fhir_ph_organisation_type": "bus",
        "fhir_address_municipality_extension_system": "StructureDefinition/address-municipality",
        "fhir_location_reference_extension_system": "StructureDefinition/address-location-reference",
        "fhir_ph_organisation_legal_form_extension_system": "StructureDefinition/organization-ph-legal-form",
        "fhir_ph_organisation_activity_extension_system": "StructureDefinition/organization-ph-activity",
        "fhir_ph_organisation_legal_form_code_system": "CodeSystem/organization-ph-legal-form",
        "fhir_ph_organisation_activity_code_system": "CodeSystem/organization-ph-activity",
        "fhir_ph_organisation_type_system": "http://terminology.hl7.org/CodeSystem/organization-type",
        "fhir_ph_organisation_type_contactentity_type":"http://terminology.hl7.org/CodeSystem/contactentity-type"
    },
    "R4_fhir_insurance_organisation_config": {
        "resource_type": 'insurance_organisation',
        "id": 'openIMIS-Implementation',
        "name": 'openIMIS Implementation',
        "code": 'openIMIS',
        "type": 'ins',
        "email": 'implementation@openimis.org',
        "phone": '0908060703',
        "fax": '0908060730',
        "contact_name": "Manuel D. Medina",
        "municipality": "Jamu",
        "city": "Jamula",
        "district": "Rapta",
        "state": "Ultha",
        "line": "1 Pasay"
    },
    "R4_fhir_invoice_config": {
        "fhir_invoice_type_system": "CodeSystem/invoice-type",
        "fhir_invoice_charge_item_system": "CodeSystem/invoice-charge-item",
        "fhir_bill_type_system": "CodeSystem/bill-type",
        "fhir_bill_charge_item_system": "CodeSystem/bill-charge-item"
    },
    "R4_fhir_subscription_config": {
        "fhir_sub_search_perms": ['158001'],
        "fhir_sub_create_perms": ['158002'],
        "fhir_sub_update_perms": ['158003'],
        "fhir_sub_delete_perms": ['158004'],
        "fhir_sub_channel_rest_hook": "rest-hook",
        "fhir_sub_status_off": "off",
        "fhir_sub_status_active": "active",
        "get_fhir_sub_criteria_key_resource": "resource",
        "get_fhir_sub_criteria_key_resource_type": "resource_type"
    }
}



from copy import deepcopy

from drf_spectacular.utils import inline_serializer
from rest_framework import fields
from rest_framework.exceptions import ValidationError, APIException, PermissionDenied

from api_fhir_r4.configurations import R4SubscriptionConfig
from api_fhir_r4.subscriptions import SubscriptionConverter
from api_fhir_r4.models import Subscription
from api_fhir_r4.permissions import FHIRApiInsureePermissions, FHIRApiInvoicePermissions, \
    FHIRApiHealthServicePermissions
from api_fhir_r4.serializers import BaseFHIRSerializer
from api_fhir_r4.services import SubscriptionService


class SubscriptionSerializer(BaseFHIRSerializer):
    fhirConverter = SubscriptionConverter
    _error_while_saving = 'Error while saving a subscription: %(msg)s'

    resource_permissions = {
        'patient': FHIRApiInsureePermissions.permissions_get,
        'invoice': FHIRApiInvoicePermissions.permissions_get,
        'organisation': FHIRApiHealthServicePermissions.permissions_get
    }

    def create(self, validated_data):
        user = self.context['request'].user
        self.check_resource_rights(user, validated_data)
        service = SubscriptionService(user)
        copied_data = deepcopy(validated_data)
        del copied_data['_state'], copied_data['_original_state']
        result = service.create(copied_data)
        return self.get_result_object(result)

    def update(self, instance, validated_data):
        user = self.context['request'].user
        self.check_instance_id(instance, validated_data)
        self.check_object_owner(user, instance)
        self.check_resource_rights(user, validated_data)
        service = SubscriptionService(user)
        copied_data = {key: value for key, value in deepcopy(validated_data).items() if value is not None}
        copied_data['id'] = instance.id
        del copied_data['_state'], copied_data['_original_state']
        result = service.update(copied_data)
        return self.get_result_object(result)

    def get_result_object(self, result):
        if result.get('success', False):
            return Subscription.objects.get(id=result['data']['id'])
        else:
            raise APIException(self._error_while_saving % {'msg': result.get('message', 'Unknown')})

    def check_resource_rights(self, user, data):
        resource = data.get('criteria', {}).get(R4SubscriptionConfig.get_fhir_sub_criteria_key_resource(), '').lower()
        if not resource or resource not in self.resource_permissions:
            raise ValidationError(self._error_while_saving % {'msg': f'Invalid resource ({resource})'})

        if not user.has_perms(self.resource_permissions[resource]):
            raise PermissionDenied(
                detail=self._error_while_saving % {'msg': f'You have no permissions to subscribe to {resource}'})

    def check_object_owner(self, user, instance):
        if str(user.id).lower() != str(instance.user_created.id).lower():
            raise APIException(self._error_while_saving % {'msg': 'You are not the owner of this subscription'})

    def check_instance_id(self, instance, validated_data):
        if validated_data['id'] and str(instance.id) != validated_data['id']:
            raise APIException(self._error_while_saving % {'msg': 'Invalid ID in the payload'})


class SubscriptionSerializerSchema(SubscriptionSerializer):
    resourceType = fields.CharField(read_only=True)
    id = fields.UUIDField(read_only=True)
    status = fields.CharField()
    end = fields.DateTimeField()
    reason = fields.CharField()
    criteria = fields.CharField()
    channel = inline_serializer(
        name='ChannelSerializer',
        fields={
            'type': fields.CharField(),
            'endpoint': fields.CharField(),
            'header': fields.ListField(child=fields.CharField(), min_length=0, max_length=1),
        },
    )



from api_fhir_r4.models.fhirModels import OperationOutcomeV2, UsageContextV2, CoverageV2, CoverageClassV2, \
    ContractSignerV2, ClaimV2, ClaimInsuranceV2, ClaimResponseV2, CoverageEligibilityRequestV2
from api_fhir_r4.models.imisModelEnums import BundleType
from api_fhir_r4.models.subscription import Subscription, SubscriptionNotificationResult



from location.models import Location

from api_fhir_r4.configurations import R4IdentifierConfig, R4LocationConfig
from api_fhir_r4.converters import LocationConverter
from fhir.resources.location import Location as FHIRLocation
from api_fhir_r4.tests import GenericTestMixin


class LocationTestMixin(GenericTestMixin):

    _TEST_CODE = "RTDTMTVT"
    _TEST_NAME = "TEST_NAME"
    _TEST_LOCATION_TYPE = "V"
    _TEST_MUNICIPALITY_UUID = "a82f54bf-d983-4963-a279-490312a96344"

    def create_test_imis_instance(self):
        # create level location
        imis_location_region = Location()
        imis_location_region.code = "RT"
        imis_location_region.name = self._TEST_NAME
        imis_location_region.type = "R"
        imis_location_region.save()

        imis_location_district = Location()
        imis_location_district.code = "RTDT"
        imis_location_district.name = self._TEST_NAME
        imis_location_district.type = "D"
        imis_location_district.parent = imis_location_region
        imis_location_district.save()

        imis_location_municipality = Location()
        imis_location_municipality.code = "RTDTMT"
        imis_location_municipality.name = self._TEST_NAME
        imis_location_municipality.type = "M"
        imis_location_municipality.parent = imis_location_district
        imis_location_municipality.uuid = self._TEST_MUNICIPALITY_UUID
        imis_location_municipality.save()

        imis_location = Location()
        imis_location.code = self._TEST_CODE
        imis_location.name = self._TEST_NAME
        imis_location.type = self._TEST_LOCATION_TYPE
        imis_location.parent = imis_location_municipality
        return imis_location

    def verify_imis_instance(self, imis_obj):
        self.assertEqual(self._TEST_CODE, imis_obj.code)
        self.assertEqual(self._TEST_NAME, imis_obj.name)
        self.assertEqual(self._TEST_LOCATION_TYPE, imis_obj.type)

    def create_test_fhir_instance(self):
        fhir_location = FHIRLocation.construct()
        identifier = LocationConverter.build_fhir_identifier(self._TEST_CODE,
                                                             R4IdentifierConfig.get_fhir_identifier_type_system(),
                                                             R4IdentifierConfig.get_fhir_location_code_type())
        fhir_location.identifier = [identifier]

        fhir_location.name = self._TEST_NAME

        system_definition = LocationConverter.PHYSICAL_TYPES.get(self._TEST_LOCATION_TYPE)
        fhir_location.physicalType = LocationConverter.build_codeable_concept(**system_definition)

        fhir_location.mode = 'instance'
        fhir_location.status = R4LocationConfig.get_fhir_code_for_active()

        return fhir_location

    def verify_fhir_instance(self, fhir_obj):
        for identifier in fhir_obj.identifier:
            code = LocationConverter.get_first_coding_from_codeable_concept(identifier.type).code
            if code == R4IdentifierConfig.get_fhir_uuid_type_code():
                self.assertEqual(fhir_obj.id, identifier.value)
            elif code == R4IdentifierConfig.get_fhir_facility_id_type():
                self.assertEqual(self._TEST_CODE, identifier.value)
        self.assertEqual(self._TEST_NAME, fhir_obj.name)
        physical_type_code = LocationConverter.get_first_coding_from_codeable_concept(fhir_obj.physicalType).code
        self.assertEqual(self._TEST_LOCATION_TYPE, physical_type_code)
        self.assertEqual(R4LocationConfig.get_fhir_code_for_active(), fhir_obj.status)
        self.assertEqual('instance', fhir_obj.mode)



from api_fhir_r4.configurations import R4IdentifierConfig
from api_fhir_r4.converters import ClaimAdminPractitionerRoleConverter
from fhir.resources.identifier import Identifier
from fhir.resources.practitionerrole import PractitionerRole
from fhir.resources.reference import Reference
from api_fhir_r4.tests import GenericTestMixin, ClaimAdminPractitionerTestMixin, LocationTestMixin
from location.models import HealthFacility


class ClaimAdminPractitionerRoleTestMixin(GenericTestMixin):

    _TEST_CLAIM_ADMIN = None
    _TEST_HF = None
    _TEST_ORGANIZATION_REFERENCE = None
    _TEST_PRACTITIONER_REFERENCE = None

    _TEST_ID = 1
    _TEST_UUID = "254f6268-964b-4d8d-aa26-20081f22235e"
    _TEST_CODE = "1234abcd"

    _TEST_HF_ID = 10000
    _TEST_HF_UUID = "6d0eea8c-62eb-11ea-94d6-c36229a16c2f"
    _TEST_HF_CODE = "12345678"
    _TEST_HF_NAME = "TEST_NAME"
    _TEST_HF_LEVEL = "H"
    _TEST_HF_LEGAL_FORM = "G"
    _TEST_ADDRESS = "TEST_ADDRESS"
    _TEST_PHONE = "133-996-476"
    _TEST_FAX = "1-408-999 8888"
    _TEST_EMAIL = "TEST@TEST.com"

    def setUp(self):
        self._TEST_CLAIM_ADMIN = ClaimAdminPractitionerTestMixin().create_test_imis_instance()
        self._TEST_CLAIM_ADMIN.save()
        self._TEST_PRACTITIONER_REFERENCE = "Practitioner/" + self._TEST_CLAIM_ADMIN.uuid
        self._TEST_HF = self.create_test_health_facility()
        self._TEST_ORGANIZATION_REFERENCE = "Organization/" + self._TEST_HF.uuid

    def create_test_health_facility(self):
        location = LocationTestMixin().create_test_imis_instance()
        location.save()
        hf = HealthFacility()
        hf.id = self._TEST_HF_ID
        hf.uuid = self._TEST_HF_UUID
        hf.code = self._TEST_HF_CODE
        hf.name = self._TEST_HF_NAME
        hf.level = self._TEST_HF_LEVEL
        hf.legal_form_id = self._TEST_HF_LEGAL_FORM
        hf.address = self._TEST_ADDRESS
        hf.phone = self._TEST_PHONE
        hf.fax = self._TEST_FAX
        hf.email = self._TEST_EMAIL
        hf.location_id = location.id
        hf.offline = False
        hf.audit_user_id = -1
        hf.save()
        return hf

    def create_test_imis_instance(self):
        self._TEST_CLAIM_ADMIN.health_facility = self._TEST_HF
        return self._TEST_CLAIM_ADMIN

    def verify_imis_instance(self, imis_obj):
        self.assertEqual(self._TEST_HF.code, imis_obj.health_facility.code)

    def create_test_fhir_instance(self):
        fhir_practitioner_role = PractitionerRole.construct()
        identifiers = []
        code = ClaimAdminPractitionerRoleConverter.build_fhir_identifier(
            self._TEST_CODE,
            R4IdentifierConfig.get_fhir_identifier_type_system(),
            R4IdentifierConfig.get_fhir_generic_type_code()
        )
        identifiers.append(code)
        fhir_practitioner_role.identifier = identifiers
        organization_reference = Reference.construct()
        organization_reference.reference = self._TEST_ORGANIZATION_REFERENCE
        fhir_practitioner_role.organization = organization_reference
        practitioner_reference = Reference.construct()
        practitioner_reference.reference = self._TEST_PRACTITIONER_REFERENCE
        fhir_practitioner_role.practitioner = practitioner_reference
        return fhir_practitioner_role

    def verify_fhir_instance(self, fhir_obj):
        self.assertIn(self._TEST_ORGANIZATION_REFERENCE, fhir_obj.organization.reference)
        for identifier in fhir_obj.identifier:
            self.assertTrue(isinstance(identifier, Identifier))
            code = ClaimAdminPractitionerRoleConverter.get_first_coding_from_codeable_concept(identifier.type).code
            if code == R4IdentifierConfig.get_fhir_generic_type_code():
                self.assertEqual(self._TEST_CODE, identifier.value)
            elif code == R4IdentifierConfig.get_fhir_uuid_type_code():
                self.assertEqual(self._TEST_UUID, identifier.value)
        self.assertIn(self._TEST_PRACTITIONER_REFERENCE, fhir_obj.practitioner.reference)
        self.assertEqual(1, len(fhir_obj.code))
        self.assertEqual(1, len(fhir_obj.code[0].coding))
        self.assertEqual("CA", fhir_obj.code[0].coding[0].code)
        self.assertEqual("Claim Administrator",  fhir_obj.code[0].coding[0].display)



from api_fhir_r4.configurations import R4CoverageConfig


class CoverageStatus(object):

    @classmethod
    def map_status(cls, code):
        codes = {
            1: R4CoverageConfig.get_status_draft_code(),
            2: R4CoverageConfig.get_status_active_code(),
            4: R4CoverageConfig.get_status_cancelled_code(),
            8: R4CoverageConfig.get_status_active_code(),
            None: R4CoverageConfig.get_status_draft_code(),
        }
        return codes[code]



from api_fhir_r4.configurations import R4IssueTypeConfig
from api_fhir_r4.converters import BaseFHIRConverter, OperationOutcomeConverter
from api_fhir_r4.exceptions import FHIRRequestProcessException
from fhir.resources.codeableconcept import CodeableConcept
from api_fhir_r4.tests import GenericTestMixin


class OperationOutcomeTestMixin(GenericTestMixin):

    __ERROR_MESSAGE = "Error message"
    __VALID_CONDITION = 1 == 1

    def create_test_imis_instance(self):
        errors = []
        BaseFHIRConverter.valid_condition(self.__VALID_CONDITION, self.__ERROR_MESSAGE, errors)
        return FHIRRequestProcessException(errors)

    def create_test_fhir_instance(self):
        exc = self.create_test_imis_instance()
        return OperationOutcomeConverter.to_fhir_obj(exc)

    def verify_fhir_instance(self, fhir_obj):
        issues = fhir_obj.issue
        self.assertEqual(1, len(issues))
        first_issue = issues[0]
        self.assertEqual(first_issue.code, R4IssueTypeConfig.get_fhir_code_for_exception())
        self.assertEqual(first_issue.severity, "error")
        details = first_issue.details
        self.assertTrue(isinstance(details, CodeableConcept))
        self.assertEqual(details.text, "The request cannot be processed due to the following issues:\nError message")



import importlib


class FunctionUtils(object):

    @classmethod
    def get_function_by_str(cls, function_string):
        mod_name, func_name = function_string.rsplit('.', 1)
        mod = importlib.import_module(mod_name)
        func = getattr(mod, func_name)
        return func



import hashlib
import urllib
from api_fhir_r4.configurations import GeneralConfiguration
from fhir.resources.bundle import Bundle, BundleEntry, BundleLink
from rest_framework.pagination import PageNumberPagination
from rest_framework.response import Response
from django.core.cache import caches
from django.db.models.query import QuerySet


class FhirBundleResultsSetPagination(PageNumberPagination):

    page_size = GeneralConfiguration.get_default_response_page_size()
    page_query_param = 'page-offset'
    page_size_query_param = '_count'

    def get_paginated_response(self, data):
        return Response(self.build_bundle_set(data).dict())

    def build_bundle_set(self, data):
        bundle = Bundle.construct()
        bundle.type = "searchset"
        bundle.total = self.page.paginator.count
        self.build_bundle_links(bundle)
        self.build_bundle_entry(bundle, data)
        return bundle

    def build_bundle_links(self, bundle):
        self.build_bundle_link(bundle, "self", self.request.build_absolute_uri())
        next_link = self.get_next_link()
        if next_link:
            self.build_bundle_link(bundle, "next", next_link)
        previous_link = self.get_previous_link()
        if previous_link:
            self.build_bundle_link(bundle, "previous", previous_link)

    def build_bundle_link(self, bundle, relation, url):
        self_link = {}
        self_link['url'] = urllib.parse.quote_plus(url)
        self_link['relation'] = relation
        bundle_link = BundleLink(**self_link)
        if type(bundle.link) is not list:
           bundle.link = [bundle_link]
        else:
           bundle.link.append(bundle_link)

    def build_bundle_entry(self, bundle, data):
        bundle.entry = []
        for obj in data:
            entry = {}
            entry['fullUrl'] = self.build_full_url_for_resource(obj)
            entry['resource'] = obj
            bundle_entry = BundleEntry(**entry)
            bundle.entry.append(bundle_entry)

    def build_full_url_for_resource(self, fhir_object):
        url = None
        resource_pk = self.get_object_pk(fhir_object)
        if resource_pk:
            url = self.request.build_absolute_uri()
            url = self.exclude_query_parameter_from_url(url)
            url = url + resource_pk
        return url

    def get_object_pk(self, fhir_object):
        pk_id = None
        if isinstance(fhir_object, dict):
            pk_id = fhir_object.get('id')
        return str(pk_id) if pk_id else None

    def exclude_query_parameter_from_url(self, url):
        try:
            from urllib.parse import urlparse  # Python 3
        except ImportError:
            from urlparse import urlparse  # Python 2
        o = urlparse(url)
        return o._replace(query=None).geturl()

    def paginate_queryset(self, queryset, *args, **kwargs):
        if isinstance(queryset, QuerySet) and hasattr(queryset, 'count'):
            queryset = CachedCountQueryset(queryset)
        return super().paginate_queryset(queryset, *args, **kwargs)


def CachedCountQueryset(queryset, timeout=60*60, cache_name='default'):
    """
        Return copy of queryset with queryset.count() wrapped to cache result for `timeout` seconds.
    """
    cache = caches[cache_name]
    queryset = queryset._chain()
    real_count = queryset.count

    def count(queryset):
        cache_key = 'query-count:' + hashlib.md5(str(queryset.query).encode('utf8')).hexdigest()

        # return existing value, if any
        value = cache.get(cache_key)
        if value is not None:
            return value

        # cache new value
        value = real_count()
        cache.set(cache_key, value, timeout)
        return value

    queryset.count = count.__get__(queryset, type(queryset))
    return queryset


import core
import uuid

from policy.test_helpers import create_test_policy
from insuree.models import InsureePolicy

from api_fhir_r4.configurations import R4IdentifierConfig
from api_fhir_r4.converters.coverageConverter import CoverageConverter
from api_fhir_r4.tests import GenericTestMixin

from api_fhir_r4.utils import TimeUtils

from core.test_helpers import create_test_officer
from insuree.test_helpers import create_test_insuree
from product.test_helpers import create_test_product


class CoverageTestMixin(GenericTestMixin):

    _TEST_POLICY_UUID = "f88687a7-1f33-466b-8c74-8b7173dc5583"
    _TEST_POLICY_ENROLL_DATE = "2021-08-20"
    _TEST_POLICY_START_DATE = "2021-08-20T00:00:00"
    _TEST_POLICY_EFFECTIVE_DATE = "2021-08-20"
    _TEST_POLICY_EXPIRED_DATE = "2022-08-19T00:00:00"
    _TEST_POLICY_STATUS = 1
    _TEST_POLICY_STAGE = 'N'
    _TEST_PRODUCT_CODE = "T0001"
    _TEST_PRODUCT_NAME = "Test product"
    _TEST_INSUREE_CHF_ID = 'chfid1'
    _TEST_PRODUCT_UUID = "8ed8d2d9-2644-4d29-ba37-ab772386cfca"

    _TEST_POLICY_VALUE = 10000.0

    def create_test_imis_instance(self):
        # create mocked insuree
        imis_insuree = create_test_insuree(
            with_family=True,
            custom_props={"chf_id": self._TEST_INSUREE_CHF_ID}
        )

        # update family uuid
        imis_family = imis_insuree.family
        # create mocked product

        imis_officer = create_test_officer()
        imis_product = create_test_product(self._TEST_PRODUCT_CODE, valid=True, custom_props=None)
        imis_product.uuid = self._TEST_PRODUCT_UUID
        imis_product.save()

        imis_policy = create_test_policy(
            product=imis_product,
            insuree=imis_insuree,
            custom_props={
                "uuid": self._TEST_POLICY_UUID,
                "officer_id": imis_officer.id
            }
        )

        imis_policy.family = imis_family
        imis_policy.product = imis_product

        imis_policy.enroll_date = TimeUtils.str_to_date(self._TEST_POLICY_ENROLL_DATE)
        imis_policy.start_date = TimeUtils.str_to_date(self._TEST_POLICY_START_DATE)
        imis_policy.effective_date = TimeUtils.str_to_date(self._TEST_POLICY_EFFECTIVE_DATE)
        imis_policy.expiry_date = TimeUtils.str_to_date(self._TEST_POLICY_EXPIRED_DATE)

        imis_policy.stage = self._TEST_POLICY_STAGE
        imis_policy.status = self._TEST_POLICY_STATUS
        imis_policy.value = self._TEST_POLICY_VALUE
        imis_policy.audit_user_id = -1

        # save mock policy
        imis_policy.save()

        # create mock policy insuree
        imis_policy_insuree = InsureePolicy(
            policy=imis_policy,
            insuree=imis_insuree,
            audit_user_id=-1
        )
        imis_policy_insuree.save()

        return imis_policy

    def verify_imis_instance(self, imis_obj):
        self.assertEqual(self._TEST_POLICY_ENROLL_DATE, imis_obj.enroll_date.isoformat())
        self.assertEqual(self._TEST_POLICY_START_DATE, imis_obj.start_date.isoformat())
        self.assertEqual(self._TEST_POLICY_EFFECTIVE_DATE, imis_obj.effective_date.isoformat())
        self.assertEqual(self._TEST_POLICY_EXPIRED_DATE, imis_obj.expiry_date.isoformat())
        self.assertEqual(self._TEST_PRODUCT_CODE, imis_obj.product.code)
        self.assertEqual(self._TEST_PRODUCT_UUID, str(uuid.UUID(imis_obj.product.uuid)))

    def verify_fhir_instance(self, fhir_obj):
        for identifier in fhir_obj.identifier:
            code = CoverageConverter.get_first_coding_from_codeable_concept(identifier.type).code
            if code == R4IdentifierConfig.get_fhir_uuid_type_code():
                self.assertEqual(self._TEST_POLICY_UUID, identifier.value)
        self.assertIn(f"Patient/{self._TEST_INSUREE_CHF_ID}", fhir_obj.policyHolder.reference)
        self.assertIn(f"Patient/{self._TEST_INSUREE_CHF_ID}", fhir_obj.beneficiary.reference)
        self.assertIn(f"Patient/{self._TEST_INSUREE_CHF_ID}", fhir_obj.payor[0].reference)
        self.assertEqual(1, len(fhir_obj.class_fhir))
        self.assertEqual('plan', fhir_obj.class_fhir[0].type.coding[0].code)
        self.assertEqual(self._TEST_PRODUCT_CODE, fhir_obj.class_fhir[0].value)
        self.assertEqual(f'{self._TEST_PRODUCT_NAME} {self._TEST_PRODUCT_CODE}', fhir_obj.class_fhir[0].name)
        period = fhir_obj.period
        self.assertEqual(self._TEST_POLICY_START_DATE, period.start.isoformat())
        self.assertEqual(self._TEST_POLICY_EXPIRED_DATE, period.end.isoformat())



from claim.models import Claim


class FeedbackStatus(object):
    
    @classmethod
    def map_status(cls, code):
        codes = {
            Claim.FEEDBACK_SELECTED: "active",
            Claim.FEEDBACK_DELIVERED: "completed",
            Claim.FEEDBACK_BYPASSED: "revoked",
        }
        return codes[code]

    @classmethod
    def map_code_display(cls, code):
        codes = {
            Claim.FEEDBACK_SELECTED: "Selected",
            Claim.FEEDBACK_DELIVERED: "Delivered",
            Claim.FEEDBACK_BYPASSED: "Bypassed",
        }
        return codes[code]



from abc import ABC, abstractmethod
from datetime import timedelta

from django.db.models import QuerySet
from typing import Dict, Callable, Any

from core.datetimes.ad_datetime import datetime


class QuerysetFilterABC(ABC):
    def __init__(self, field, value):
        self.field = field
        self.value = value

    @abstractmethod
    def apply_filter(self, queryset: QuerySet) -> QuerySet:
        """
        apply_filter should apply contained filter to given queryset and return the result.
        @param queryset: queryset the filter should be applied to
        @return: filtered queryset
        """
        pass


class QuerysetEqualFilter(QuerysetFilterABC):
    def apply_filter(self, queryset):
        return queryset.filter(**{self.field: self.value})


class QuerysetNotEqualFilter(QuerysetFilterABC):
    def apply_filter(self, queryset):
        return queryset.exclude(**{self.field: self.value})


class QuerysetGreaterThanFilter(QuerysetFilterABC):
    def apply_filter(self, queryset):
        return queryset.filter(**{f'{self.field}__gt': self.value})


class QuerysetLesserThanFilter(QuerysetFilterABC):
    def apply_filter(self, queryset):
        return queryset.filter(**{f'{self.field}__lt': self.value})


class QuerysetGreaterThanEqualFilter(QuerysetFilterABC):
    def apply_filter(self, queryset):
        return queryset.filter(**{f'{self.field}__gte': self.value})


class QuerysetLesserThanEqualFilter(QuerysetFilterABC):
    def apply_filter(self, queryset):
        return queryset.filter(**{f'{self.field}__lte': self.value})


class QuerysetApproximateDateFilter(QuerysetFilterABC):
    def apply_filter(self, queryset):
        range_size = (datetime.now() - self.value).days * 0.1
        value_start = self.value - timedelta(days=range_size)
        value_end = self.value + timedelta(days=range_size)
        return queryset.filter(**{f'{self.field}__range': (value_start, value_end)})


class QuerysetParameterABC(ABC):
    def __init__(self, output_parameter):
        self.output_parameter = output_parameter
        self.accepted_prefixes = self._get_prefix_filter_mapping().keys()

    @abstractmethod
    def _get_prefix_filter_mapping(self) -> Dict[str, Callable[[str, Any], QuerysetFilterABC]]:
        """
        _get_prefix_filter_mapping should return a dict that maps respective prefixes from FHIR specification to lambdas
        capable of creating specific filters taking affected field and parsed parameter value as arguments.
        @return: {prefix: lambda creating filter} map
        """
        pass

    def build_filter(self, request_parameter_value):
        modifier, value = self._get_prefix_and_value(request_parameter_value)
        return self._get_prefix_filter_mapping()[modifier if modifier else 'eq'](self.output_parameter, value)

    def _get_prefix_and_value(self, parameter):
        modifier = next((modifier for modifier in self.accepted_prefixes if parameter.startswith(modifier)), '')
        output_value = self._parse_value(parameter[len(modifier):])
        return modifier, output_value

    def _parse_value(self, value):
        """
        Allow for custom parameter value parsing logic. _parse_value in case of parsing error should raise Value error
        with message containing {request_parameter} placeholder to insert parameter name
        @param value: value to be parsed
        @return: parsed value
        """
        return value


class QuerysetLastUpdatedParameter(QuerysetParameterABC):
    def _get_prefix_filter_mapping(self):
        return {
            'eq': lambda field, value: QuerysetEqualFilter(field, value),
            'ne': lambda field, value: QuerysetNotEqualFilter(field, value),
            'gt': lambda field, value: QuerysetGreaterThanFilter(field, value),
            'lt': lambda field, value: QuerysetLesserThanFilter(field, value),
            'ge': lambda field, value: QuerysetGreaterThanEqualFilter(field, value),
            'le': lambda field, value: QuerysetLesserThanEqualFilter(field, value),
            'sa': lambda field, value: QuerysetGreaterThanEqualFilter(field, value),
            'eb': lambda field, value: QuerysetLesserThanEqualFilter(field, value),
            'ap': lambda field, value: QuerysetApproximateDateFilter(field, value)
        }

    def _parse_value(self, value):
        try:
            return datetime.strptime(value, '%Y-%m-%dT%H:%M:%S')
        except Exception:
            raise ValueError('{request_parameter} value is not a valid datetime')


class RequestParameterFilterABC(ABC):
    def __init__(self, request):
        self.request = request

    @abstractmethod
    def _get_parameter_mapping(self) -> Dict[str, Callable[[], QuerysetParameterABC]]:
        """
        _get_parameter_mapping should return a dict mapping request parameters to lambdas capable of creating
        filters (allowing lazy loading)
        @return: {request parameter: lambda creating queryset parameter} map
        """
        pass

    def filter_queryset(self, queryset):
        parameter_mapping = self._get_parameter_mapping()
        accepted_parameters = parameter_mapping.keys()
        request_parameters = {parameter: self.request.GET[parameter]
                              for parameter in accepted_parameters if parameter in self.request.GET}

        output_queryset = queryset
        for request_parameter in request_parameters:
            output_parameter = parameter_mapping[request_parameter]()
            try:
                output_queryset = output_parameter.build_filter(self.request.GET[request_parameter]).apply_filter(
                    output_queryset)
            except ValueError as parsingError:
                raise ValueError(str(parsingError).format(**{'request_parameter': request_parameter}))

        return output_queryset


class ValidityFromRequestParameterFilter(RequestParameterFilterABC):
    def _get_parameter_mapping(self):
        return {
            '_lastUpdated': lambda: QuerysetLastUpdatedParameter('validity_from'),
        }


class DateUpdatedRequestParameterFilter(RequestParameterFilterABC):
    def _get_parameter_mapping(self):
        return {
            '_lastUpdated': lambda: QuerysetLastUpdatedParameter('date_updated'),
        }



from drf_spectacular.utils import extend_schema, extend_schema_view
from graphql_jwt.utils import jwt_payload

from api_fhir_r4.openapi_schema_extensions import get_inline_login_request_serializer, \
    get_inline_login_200_response_serializer
from core.jwt import *
from core.models import User
from rest_framework import viewsets
from rest_framework.permissions import AllowAny
from rest_framework.response import Response


@extend_schema_view(
    create=extend_schema(
        request=get_inline_login_request_serializer(),
        responses={
            (200, 'application/json'): get_inline_login_200_response_serializer(),
            (401,): None
        }
    )
)
class LoginView(viewsets.ViewSet):
    permission_classes = (AllowAny,)

    def create(self, request, *args, **kwargs):
        data = request.data
        # check if we have both required data in request payload
        if 'username' in data and 'password' in data:
            username = data['username']
            password = data['password']
            users = User.objects.filter(username=username)
            # check if user with such username exists
            if len(list(users)) == 1:
                # validate provided password from payload
                user = users.first()
                if user.check_password(password):
                    # set the user to context
                    request.user = user
                    # take the payload base on user data - using same mechanism as
                    # in graphql_jwt with generating payload.
                    payload = jwt_payload(user=user)
                    # encode token based on payload
                    token = jwt_encode_user_key(payload=payload, context=request)
                    if token:
                        # return ok
                        response = {
                            "token": token,
                            "exp": payload['exp'],
                        }
                        return Response(data=response, status=200)
            # return unauthorized
            return Response(status=401)
        else:
            # return bad request
            return Response(status=400)



from api_fhir_r4.converters.contractConverter import ContractConverter
from api_fhir_r4.serializers import BaseFHIRSerializer
from contribution.gql_mutations import update_or_create_premium
from policy.services import PolicyService
from policy.models import Policy
import copy
from api_fhir_r4.exceptions import FHIRException


class ContractSerializer(BaseFHIRSerializer):
    fhirConverter = ContractConverter

    def create(self, validated_data):
        request = self.context.get("request")
        user = request.user

        family = validated_data.get('family_id')
        insurees = validated_data.pop('insurees')
        premiums = validated_data.pop('contributions')

        if Policy.objects.filter(family_id=family).filter(start_date__range=[validated_data.get('effective_date'),validated_data.get('expiry_date')]).count() > 0:
            raise FHIRException('Contract exists for this patient')

        copied_data = copy.deepcopy(validated_data)
        del copied_data['_state']

        new_policy = PolicyService(user).update_or_create(copied_data, user)
        # create contributions related to newly created policy
        if premiums:
            for premium in premiums:
                premium = premium.__dict__
                del premium['_state']
                premium['policy_uuid'] = new_policy.uuid
                premium['policy_id'] = new_policy.id
                premium['audit_user_id'] = copied_data['audit_user_id']
                premium = update_or_create_premium(premium, user)
                new_policy = premium.policy

        return new_policy
    
    def update(self, instance, validated_data):
        instance.save()
        return instance



from fhir.resources.annotation import Annotation
from fhir.resources.extension import Extension
from api_fhir_r4.configurations import GeneralConfiguration, R4IdentifierConfig
from api_fhir_r4.converters import BaseFHIRConverter, ReferenceConverterMixin
from fhir.resources.invoice import Invoice as FHIRInvoice, \
    InvoiceLineItem as FHIRInvoiceLineItem, InvoiceLineItemPriceComponent
from fhir.resources.money import Money
from api_fhir_r4.mapping.invoiceMapping import InvoiceChargeItemMapping, InvoiceTypeMapping, BillTypeMapping, \
    BillChargeItemMapping
from api_fhir_r4.utils import DbManagerUtils
from insuree.models import Insuree
from invoice.models import Invoice


class GenericInvoiceConverter(BaseFHIRConverter, ReferenceConverterMixin):
    @classmethod
    def to_imis_obj(cls, data, audit_user_id):
        raise NotImplementedError('to_imis_obj() not implemented.')

    @classmethod
    def to_fhir_obj(cls, imis_invoice, reference_type=ReferenceConverterMixin.UUID_REFERENCE_TYPE):
        fhir_invoice = {"status": "active"}
        fhir_invoice = FHIRInvoice(**fhir_invoice)
        cls.build_fhir_identifiers(fhir_invoice, imis_invoice)
        cls.build_fhir_pk(fhir_invoice, imis_invoice, reference_type)
        cls.build_fhir_type(fhir_invoice, imis_invoice)
        cls.build_fhir_recipient(fhir_invoice, imis_invoice, reference_type)
        cls.build_fhir_date(fhir_invoice, imis_invoice)
        cls.build_fhir_totals(fhir_invoice, imis_invoice)
        cls.build_fhir_line_items(fhir_invoice, imis_invoice, imis_invoice.currency_code)
        cls.build_fhir_note(fhir_invoice, imis_invoice)
        return fhir_invoice

    @classmethod
    def get_reference_obj_uuid(cls, imis_invoice):
        return imis_invoice.uuid

    @classmethod
    def get_reference_obj_id(cls, imis_invoice):
        return imis_invoice.uuid

    @classmethod
    def get_reference_obj_code(cls, imis_invoice):
        return imis_invoice.code

    @classmethod
    def get_fhir_resource_type(cls):
        return FHIRInvoice

    @classmethod
    def get_imis_obj_by_fhir_reference(cls, reference, errors=None):
        invoice_uuid = cls.get_resource_id_from_reference(reference)
        return DbManagerUtils.get_object_or_none(Invoice, uuid=invoice_uuid)

    @classmethod
    def get_fhir_code_identifier_type(cls):
        return R4IdentifierConfig.get_fhir_generic_type_code()

    @classmethod
    def build_fhir_identifiers(cls, fhir_invoice, imis_invoice):
        identifiers = []
        cls.build_fhir_uuid_identifier(identifiers, imis_invoice)
        cls.build_fhir_code_identifier(identifiers, imis_invoice)
        fhir_invoice.identifier = identifiers

    @classmethod
    def build_fhir_type(cls, fhir_invoice, imis_invoice):
        invoice_type = cls.get_type_mapping()[imis_invoice.subject_type.model]
        fhir_invoice.type = cls.build_codeable_concept_from_coding(cls.build_fhir_mapped_coding(invoice_type))

    @classmethod
    def build_fhir_recipient(cls, fhir_invoice, imis_invoice, reference_type):
        if isinstance(imis_invoice.thirdparty, Insuree):
            reference_resource = 'Patient'
        else:
            reference_resource = 'Organization'

        fhir_invoice.recipient = cls.build_fhir_resource_reference(
            imis_invoice.thirdparty,
            type=reference_resource,
            reference_type=reference_type)

    @classmethod
    def build_fhir_date(cls, fhir_invoice, imis_invoice):
        fhir_invoice.date = cls.get_invoice_date(imis_invoice)

    @classmethod
    def build_fhir_totals(cls, fhir_invoice, imis_invoice):
        fhir_invoice.totalNet = cls.build_fhir_money(imis_invoice.amount_net, imis_invoice.currency_code)
        fhir_invoice.totalGross = cls.build_fhir_money(imis_invoice.amount_total, imis_invoice.currency_code)

    @classmethod
    def build_fhir_line_items(cls, fhir_invoice, imis_invoice, currency):
        fhir_invoice.lineItem = []
        for line in cls.get_invoice_line_items(imis_invoice):
            fhir_line_item = FHIRInvoiceLineItem.construct()
            cls.build_line_item_charge_item_codeable_concept(fhir_line_item, line)
            fhir_line_item.priceComponent = []
            cls.build_line_item_price_component_base(fhir_line_item, line, currency)
            if line.tax_rate:
                cls.build_line_item_price_component_tax(fhir_line_item, line, currency)
            if line.discount:
                cls.build_line_item_price_component_discount(fhir_line_item, line, currency)
            if line.deduction:
                cls.build_line_item_price_component_deduction(fhir_line_item, line, currency)
            fhir_invoice.lineItem.append(fhir_line_item)

    @classmethod
    def build_line_item_charge_item_codeable_concept(cls, fhir_line_item, line_item):
        charge_item = cls.get_charge_item_mapping()[line_item.line_type.model]
        codeable_concept = cls.build_codeable_concept_from_coding(cls.build_fhir_mapped_coding(charge_item))
        fhir_line_item.chargeItemCodeableConcept = codeable_concept

    @classmethod
    def build_line_item_price_component_base(cls, fhir_line_item, line_item, currency):
        price_component = cls.build_line_item_price_component(
            "base",
            line_item.quantity,
            line_item.unit_price * line_item.quantity,
            currency
        )
        price_component.code = cls.build_codeable_concept(
            code="Code",
            display=line_item.code,
            system="Code"
        )
        price_component.extension = [cls.build_line_item_unit_price_extension(line_item, currency)]
        fhir_line_item.priceComponent.append(price_component)

    @classmethod
    def build_line_item_price_component_discount(cls, fhir_line_item, line_item, currency):
        price_component = cls.build_line_item_price_component(
            "discount",
            line_item.discount,
            -line_item.unit_price * line_item.quantity * line_item.discount,
            currency
        )
        price_component.extension = [cls.build_line_item_unit_price_extension(line_item, currency)]
        fhir_line_item.priceComponent.append(price_component)

    @classmethod
    def build_line_item_price_component_deduction(cls, fhir_line_item, line_item, currency):
        price_component = cls.build_line_item_price_component(
            "deduction",
            1,
            -line_item.deduction,
            currency
        )
        price_component.extension = [cls.build_line_item_unit_price_extension(line_item, currency)]
        fhir_line_item.priceComponent.append(price_component)

    @classmethod
    def build_line_item_price_component_tax(cls, fhir_line_item, line_item, currency):
        price_component = cls.build_line_item_price_component(
            "tax",
            line_item.tax_rate,
            ((line_item.quantity * line_item.unit_price * (1 - (line_item.discount or 0))) - (
                    line_item.deduction or 0)) * line_item.tax_rate,
            currency
        )
        price_component.extension = [cls.build_line_item_unit_price_extension(line_item, currency)]
        fhir_line_item.priceComponent.append(price_component)

    @classmethod
    def build_line_item_price_component(cls, component_type, factor, amount, currency):
        price_component = {"type": component_type}
        price_component = InvoiceLineItemPriceComponent(**price_component)
        price_component.factor = factor
        price_component.amount = cls.build_fhir_money(amount, currency)
        return price_component

    @classmethod
    def build_line_item_unit_price_extension(cls, line_item, currency):
        extension = Extension.construct()
        extension.url = f"{GeneralConfiguration.get_system_base_url()}/StructureDefinition/unit-price"
        unit_price = Money.construct()
        unit_price.value = line_item.unit_price
        unit_price.currency = currency
        extension.valueMoney = unit_price
        return extension

    @classmethod
    def build_fhir_note(cls, fhir_invoice, imis_invoice):
        if imis_invoice.note:
            annotation = Annotation.construct()
            annotation.text = imis_invoice.note
            fhir_invoice.note = [annotation]

    @classmethod
    def get_type_mapping(cls):
        raise NotImplementedError('get_type_mapping() not implemented')

    @classmethod
    def get_charge_item_mapping(cls):
        raise NotImplementedError('get_charge_item_mapping() not implemented')

    @classmethod
    def get_invoice_date(cls, invoice):
        raise NotImplementedError('get_invoice_date() not implemented')

    @classmethod
    def get_invoice_line_items(cls, invoice):
        raise NotImplementedError('get_invoice_date() not implemented')


class InvoiceConverter(GenericInvoiceConverter):
    @classmethod
    def get_type_mapping(cls):
        return InvoiceTypeMapping.invoice_type

    @classmethod
    def get_charge_item_mapping(cls):
        return InvoiceChargeItemMapping.charge_item

    @classmethod
    def get_invoice_date(cls, invoice):
        return invoice.date_invoice

    @classmethod
    def get_invoice_line_items(cls, invoice):
        return invoice.line_items.all()


class BillInvoiceConverter(GenericInvoiceConverter):
    @classmethod
    def get_type_mapping(cls):
        return BillTypeMapping.invoice_type

    @classmethod
    def get_charge_item_mapping(cls):
        return BillChargeItemMapping.charge_item

    @classmethod
    def get_invoice_date(cls, invoice):
        return invoice.date_bill

    @classmethod
    def get_invoice_line_items(cls, invoice):
        return invoice.line_items_bill.all()



import uuid

from policy.services import EligibilityResponse, EligibilityRequest
from insuree.test_helpers import create_test_insuree
from medical.test_helpers import create_test_service, create_test_item


from api_fhir_r4.configurations import GeneralConfiguration, R4CoverageEligibilityConfiguration as Config
from api_fhir_r4.converters import PatientConverter, CoverageEligibilityRequestConverter
from api_fhir_r4.models import CoverageEligibilityRequestV2 as CoverageEligibilityRequest
from fhir.resources.coverageeligibilityrequest import CoverageEligibilityRequestItem
from api_fhir_r4.tests import GenericTestMixin, PatientTestMixin
from api_fhir_r4.utils import TimeUtils



class CoverageEligibilityRequestTestMixin(GenericTestMixin):

    _TEST_ADMIN_USER_ID = 1
    _TEST_ADMIN_USER_UUID = "90b0cee2-73ae-4705-af8d-8fe035209ab2"
    _TEST_SERVICE_CODE = 'STEST'
    _TEST_ITEM_CODE = 'ITEST'
    _TEST_CHFID = 'chfid'
    _TEST_TOTAL_ADMISSIONS = 1
    _TEST_TOTAL_VISITS = 3
    _TEST_TOTAL_CONSULTATIONS = 23
    _TEST_TOTAL_SURGERIES = 5
    _TEST_TOTAL_DELIVERIES = 2
    _TEST_TOTAL_ANTENATAL = 11
    _TEST_CONSULTATION_AMOUNT = 123.21
    _TEST_SURGERY_AMOUNT = 12.11
    _TEST_DELIVERY_AMOUNT = 26.00
    _TEST_HOSPITALIZATION_AMOUNT = 42.00
    _TEST_ANTENATAL_AMOUNT = 59.14
    _TEST_SERVICE_LEFT = 12
    _TEST_ITEM_LEFT = 1
    _TEST_IS_SERVICE_OK = True
    _TEST_IS_ITEM_OK = False

    def setUp(self):
        self._TEST_INSUREE = create_test_insuree()
        self._TEST_INSUREE.chf_id = self._TEST_CHFID
        self._TEST_INSUREE.save()
        self._TEST_SERVICE = create_test_service('D')
        self._TEST_SERVICE.code = self._TEST_SERVICE_CODE
        self._TEST_SERVICE.save()
        self._TEST_ITEM = create_test_item('D')
        self._TEST_ITEM.code = self._TEST_ITEM_CODE
        self._TEST_ITEM.save()
        self._TEST_ELIGIBILITY_REQUEST = EligibilityRequest(
            self._TEST_CHFID,
            self._TEST_SERVICE_CODE,
            self._TEST_ITEM_CODE
        )

    def create_test_imis_instance(self):
        return EligibilityResponse(
            eligibility_request=self._TEST_ELIGIBILITY_REQUEST,
            prod_id=None,
            total_admissions_left=self._TEST_TOTAL_ADMISSIONS,
            total_visits_left=self._TEST_TOTAL_VISITS,
            total_consultations_left=self._TEST_TOTAL_CONSULTATIONS,
            total_surgeries_left=self._TEST_TOTAL_SURGERIES,
            total_deliveries_left=self._TEST_TOTAL_DELIVERIES,
            total_antenatal_left=self._TEST_TOTAL_ANTENATAL,
            consultation_amount_left=self._TEST_CONSULTATION_AMOUNT,
            surgery_amount_left=self._TEST_SURGERY_AMOUNT,
            delivery_amount_left=self._TEST_DELIVERY_AMOUNT,
            hospitalization_amount_left=self._TEST_HOSPITALIZATION_AMOUNT,
            antenatal_amount_left=self._TEST_ANTENATAL_AMOUNT,
            min_date_service=None,
            min_date_item=None,
            service_left=self._TEST_SERVICE_LEFT,
            item_left=self._TEST_ITEM_LEFT,
            is_item_ok=self._TEST_IS_ITEM_OK,
            is_service_ok=self._TEST_IS_SERVICE_OK
        )

    def verify_imis_instance(self, imis_obj):
        # The reference used to be the CHFID but reference are now all on UUID
        # self.assertEqual(self._TEST_CHFID, imis_obj.chf_id)
        self.assertEqual(str(self._TEST_INSUREE.chf_id), imis_obj.chf_id)
        self.assertEqual(self._TEST_ITEM_CODE, imis_obj.item_code)
        self.assertEqual(self._TEST_SERVICE_CODE, imis_obj.service_code)

    def create_test_fhir_instance(self):
        self.setUp()
        fhir_request = {}
        fhir_request["resourceType"] = "CoverageEligibilityRequest"
        fhir_request["status"] = "active"
        fhir_request["purpose"] = ["benefits"]
        fhir_request["created"] = TimeUtils.date().isoformat()

        reference_patient = {}
        reference_patient["reference"] = f'Patient/{self._TEST_CHFID}'
        fhir_request['patient'] = reference_patient

        reference_insurer = {}
        reference_insurer["reference"] = 'openIMIS'
        fhir_request["insurer"] = reference_insurer

        fhir_request = CoverageEligibilityRequest(**fhir_request)

        fhir_request.item = []
        item = CoverageEligibilityRequestItem.construct()
        item.category = PatientConverter.build_codeable_concept(
            system=f'{GeneralConfiguration.get_base_url()}/CodeSystem/claim-item-category',
            code="service",
            display="Service"
        )
        item.productOrService = PatientConverter.build_simple_codeable_concept(
            text=self._TEST_SERVICE_CODE
        )
        fhir_request.item.append(item)

        item = CoverageEligibilityRequestItem.construct()
        item.category = PatientConverter.build_codeable_concept(
            system=f'{GeneralConfiguration.get_base_url()}/CodeSystem/claim-item-category',
            code="item",
            display="Item"
        )
        item.productOrService = PatientConverter.build_simple_codeable_concept(
            text=self._TEST_ITEM_CODE
        )
        fhir_request.item.append(item)
        return fhir_request

    def verify_fhir_instance(self, fhir_obj):
        self.assertIsNotNone(fhir_obj.insurance[0].item)
        for item in fhir_obj.insurance[0].item:
            if item.category.text == Config.get_fhir_total_admissions_code():
                self.assertEqual(self._TEST_TOTAL_ADMISSIONS, item.benefit[0].allowedUnsignedInt)
            elif item.category.text == Config.get_fhir_total_visits_code():
                self.assertEqual(self._TEST_TOTAL_VISITS, item.benefit[0].allowedUnsignedInt)
            elif item.category.text == Config.get_fhir_total_consultations_code():
                self.assertEqual(self._TEST_TOTAL_CONSULTATIONS, item.benefit[0].allowedUnsignedInt)
            elif item.category.text == Config.get_fhir_total_surgeries_code():
                self.assertEqual(self._TEST_TOTAL_SURGERIES, item.benefit[0].allowedUnsignedInt)
            elif item.category.text == Config.get_fhir_total_deliveries_code():
                self.assertEqual(self._TEST_TOTAL_DELIVERIES, item.benefit[0].allowedUnsignedInt)
            elif item.category.text == Config.get_fhir_total_antenatal_code():
                self.assertEqual(self._TEST_TOTAL_ANTENATAL, item.benefit[0].allowedUnsignedInt)
            elif item.category.text == Config.get_fhir_consultation_amount_code():
                self.assertEqual(str(self._TEST_CONSULTATION_AMOUNT), str(item.benefit[0].allowedMoney.value))
            elif item.category.text == Config.get_fhir_surgery_amount_code():
                self.assertEqual(str(self._TEST_SURGERY_AMOUNT), str(item.benefit[0].allowedMoney.value))
            elif item.category.text == Config.get_fhir_delivery_amount_code():
                self.assertEqual(str(self._TEST_DELIVERY_AMOUNT), str(item.benefit[0].allowedMoney.value))
            elif item.category.text == Config.get_fhir_hospitalization_amount_code():
                self.assertEqual(str(self._TEST_HOSPITALIZATION_AMOUNT), str(item.benefit[0].allowedMoney.value))
            elif item.category.text == Config.get_fhir_antenatal_amount_code():
                self.assertEqual(str(self._TEST_ANTENATAL_AMOUNT), str(item.benefit[0].allowedMoney.value))
            elif item.category.text == Config.get_fhir_service_left_code():
                self.assertEqual(self._TEST_SERVICE_LEFT, item.benefit[0].allowedUnsignedInt)
            elif item.category.text == Config.get_fhir_item_left_code():
                self.assertEqual(self._TEST_ITEM_LEFT, item.benefit[0].allowedUnsignedInt)
            elif item.category.text == Config.get_fhir_is_service_ok_code():
                self.assertEqual(self._TEST_IS_SERVICE_OK, not item.excluded)
            elif item.category.text == Config.get_fhir_is_item_ok_code():
                self.assertEqual(self._TEST_IS_ITEM_OK, not item.excluded)



import json
import os

from rest_framework import status
from rest_framework.test import APITestCase

from api_fhir_r4.configurations import GeneralConfiguration
from api_fhir_r4.tests import GenericFhirAPITestMixin
from api_fhir_r4.tests import LocationTestMixin, ClaimAdminPractitionerTestMixin
from api_fhir_r4.tests.mixin.logInMixin import LogInMixin
from api_fhir_r4.utils import TimeUtils
from claim.models import Claim
from insuree.test_helpers import create_test_insuree
from location.models import HealthFacility, UserDistrict
from medical.models import Diagnosis
from medical.test_helpers import create_test_item, create_test_service


class ClaimAPITests(GenericFhirAPITestMixin, APITestCase, LogInMixin):
    base_url = GeneralConfiguration.get_base_url() + 'Claim/'
    _test_json_path = "/test/test_claim.json"

    # diagnosis data
    _TEST_MAIN_ICD_CODE = 'T_CD'
    _TEST_MAIN_ICD_NAME = 'Test diagnosis'

    _TEST_CLAIM_ADMIN_UUID = "044c33d1-dbf3-4d6a-9924-3797b461e535"
    _TEST_INSUREE_UUID = "76aca309-f8cf-4890-8f2e-b416d78de00b"

    # claim item data
    _TEST_ITEM_CODE = "iCode"
    _TEST_ITEM_UUID = "e2bc1546-390b-4d41-8571-632ecf7a0936"
    _TEST_ITEM_QUANTITY_PROVIDED = 10.0
    _TEST_ITEM_PRICE_ASKED = 10.0
    _TEST_ITEM_EXPLANATION = "item_explanation"
    _TEST_ITEM_TYPE = 'D'

    # claim service data
    _TEST_SERVICE_CODE = "sCode"
    _TEST_SERVICE_UUID = "a17602f4-e9ff-4f42-a6a4-ccefcb23b4d6"
    _TEST_SERVICE_QUANTITY_PROVIDED = 1
    _TEST_SERVICE_PRICE_ASKED = 21000.0
    _TEST_SERVICE_EXPLANATION = "service_explanation"
    _TEST_SERVICE_TYPE = 'D'

    # hf test data
    _TEST_HF_ID = 10000
    _TEST_HF_UUID = "6d0eea8c-62eb-11ea-94d6-c36229a16c2f"
    _TEST_HF_CODE = "12345678"
    _TEST_HF_NAME = "TEST_NAME"
    _TEST_HF_LEVEL = "H"
    _TEST_HF_LEGAL_FORM = "G"
    _TEST_ADDRESS = "TEST_ADDRESS"
    _TEST_PHONE = "133-996-476"
    _TEST_FAX = "1-408-999 8888"
    _TEST_EMAIL = "TEST@TEST.com"

    _ADMIN_AUDIT_USER_ID = -1

    _test_json_path_credentials = "/tests/test/test_login.json"
    _test_request_data_credentials = None

    def setUp(self):
        super(ClaimAPITests, self).setUp()
        dir_path = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
        json_representation = open(dir_path + self._test_json_path_credentials).read()
        self._test_request_data_credentials = json.loads(json_representation)
        self._TEST_USER = self.get_or_create_user_api()
        self.create_dependencies()

    def create_dependencies(self):
        self._TEST_DIAGNOSIS_CODE = Diagnosis()
        self._TEST_DIAGNOSIS_CODE.code = self._TEST_MAIN_ICD_CODE
        self._TEST_DIAGNOSIS_CODE.name = self._TEST_MAIN_ICD_NAME
        self._TEST_DIAGNOSIS_CODE.audit_user_id = self._ADMIN_AUDIT_USER_ID
        self._TEST_DIAGNOSIS_CODE.save()

        self._TEST_CLAIM_ADMIN = ClaimAdminPractitionerTestMixin().create_test_imis_instance()
        self._TEST_CLAIM_ADMIN.uuid = self._TEST_CLAIM_ADMIN_UUID
        self._TEST_CLAIM_ADMIN.save()
        self._TEST_HF = self.create_test_health_facility()

        self._TEST_INSUREE = create_test_insuree()
        self._TEST_INSUREE.uuid = self._TEST_INSUREE_UUID
        self._TEST_INSUREE.save()
        self._TEST_ITEM = self.create_test_claim_item()
        self._TEST_SERVICE = self.create_test_claim_service()

    def create_test_health_facility(self):
        location = LocationTestMixin().create_test_imis_instance()
        location.save()
        hf = HealthFacility()
        hf.id = self._TEST_HF_ID
        hf.uuid = self._TEST_HF_UUID
        hf.code = self._TEST_HF_CODE
        hf.name = self._TEST_HF_NAME
        hf.level = self._TEST_HF_LEVEL
        hf.legal_form_id = self._TEST_HF_LEGAL_FORM
        hf.address = self._TEST_ADDRESS
        hf.phone = self._TEST_PHONE
        hf.fax = self._TEST_FAX
        hf.email = self._TEST_EMAIL
        hf.location_id = location.id
        hf.offline = False
        hf.audit_user_id = self._ADMIN_AUDIT_USER_ID
        hf.save()
        ud = UserDistrict()
        ud.location = location
        ud.audit_user_id = self._ADMIN_AUDIT_USER_ID
        ud.user = self._TEST_USER.i_user
        ud.validity_from = TimeUtils.now()
        ud.save()
        return hf

    def create_test_claim_item(self):
        item = create_test_item(
            self._TEST_ITEM_TYPE,
            custom_props={"code": self._TEST_ITEM_CODE}
        )
        item.code = self._TEST_ITEM_CODE
        item.uuid = self._TEST_ITEM_UUID
        item.save()
        return item

    def create_test_claim_service(self):
        service = create_test_service(
            self._TEST_SERVICE_TYPE,
            custom_props={"code": self._TEST_SERVICE_CODE}
        )
        service.code = self._TEST_SERVICE_CODE
        service.uuid = self._TEST_SERVICE_UUID
        service.save()
        return service

    def test_post_should_create_correctly(self):
        response = self.client.post(
            GeneralConfiguration.get_base_url() + 'login/', data=self._test_request_data_credentials, format='json'
        )
        response_json = response.json()
        token = response_json["token"]
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        headers = {
            "Content-Type": "application/json",
            "HTTP_AUTHORIZATION": f"Bearer {token}"
        }
        response = self.client.post(self.base_url, data=self._test_request_data, format='json', **headers)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertIsNotNone(response.content)
        response_json = response.json()
        self.assertEqual(response_json["resourceType"], 'ClaimResponse')
        self.assertEqual(response_json["outcome"], 'complete')
        # for both tests item and service should be 'rejected' and 'not in price list'
        for item in response_json["item"]:
            for adjudication in item["adjudication"]:
                self.assertEqual(adjudication["category"]["coding"][0]["code"], f'{Claim.STATUS_REJECTED}')
                # 2 not in price list
                self.assertEqual(adjudication["reason"]["coding"][0]["code"], '2')

        self.assertEqual(response_json["resourceType"], "ClaimResponse")

    def test_get_should_return_200_claim_response(self):
        # test if get ClaimResponse return 200
        response = self.client.post(
            GeneralConfiguration.get_base_url() + 'login/', data=self._test_request_data_credentials, format='json'
        )
        response_json = response.json()
        token = response_json["token"]
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        headers = {
            "Content-Type": "application/json",
            "HTTP_AUTHORIZATION": f"Bearer {token}"
        }
        response = self.client.get(GeneralConfiguration.get_base_url() + 'ClaimResponse/', data=None, format='json',
                                   **headers)
        self.assertEqual(response.status_code, status.HTTP_200_OK)



from drf_spectacular.utils import extend_schema, extend_schema_view
from rest_framework.exceptions import PermissionDenied, APIException
from rest_framework.viewsets import ModelViewSet

from api_fhir_r4.models import Subscription
from api_fhir_r4.openapi_schema_extensions import get_inline_error_serializer
from api_fhir_r4.permissions import FHIRApiSubscriptionPermissions
from api_fhir_r4.subscriptions import SubscriptionSerializer
from api_fhir_r4.services import SubscriptionService
from api_fhir_r4.subscriptions.subscriptionSerializer import SubscriptionSerializerSchema
from api_fhir_r4.views.fhir.base import BaseFHIRView
from api_fhir_r4.views.filters import DateUpdatedRequestParameterFilter


@extend_schema_view(
    list=extend_schema(responses={(200, 'application/json'): SubscriptionSerializerSchema()}),
    retrieve=extend_schema(responses={
        (200, 'application/json'): SubscriptionSerializerSchema(),
        (404, 'application/json'): get_inline_error_serializer()
    }),
    create=extend_schema(
        request=SubscriptionSerializerSchema(),
        responses={(201, 'application/json'): SubscriptionSerializerSchema()}
    ),
    update=extend_schema(
        request=SubscriptionSerializerSchema(),
        responses={(200, 'application/json'): SubscriptionSerializerSchema()}
    ),
    destroy=extend_schema(responses={204: None})
)
class SubscriptionViewSet(BaseFHIRView, ModelViewSet):
    _error_while_deleting = 'Error while deleting a subscription: %(msg)s'
    serializer_class = SubscriptionSerializer
    http_method_names = ('get', 'post', 'put', 'delete')
    permission_classes = (FHIRApiSubscriptionPermissions,)

    def get_queryset(self):
        queryset = Subscription.objects.filter(is_deleted=False).order_by('date_created')
        return DateUpdatedRequestParameterFilter(self.request).filter_queryset(queryset)

    def perform_destroy(self, instance):
        if not self.check_if_owner(self.request.user, instance):
            raise PermissionDenied(
                detail=self._error_while_deleting % {'msg': 'You are not the owner of this subscription'})
        service = SubscriptionService(self.request.user)
        result = service.delete({'id': instance.uuid})
        self.check_error_message(result)

    @staticmethod
    def check_error_message(result):
        if not result.get('success', False):
            msg = result.get('message', 'Unknown')
            raise APIException(f'Error while deleting a subscription: {msg}')

    @staticmethod
    def check_if_owner(user, instance):
        return str(user.id).lower() == str(instance.user_created.id).lower()
